"""Library for visualizations used in META-SiM."""
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import umap
    import numpy as np
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    import matplotlib.colors as plt_colors
    import metasim.fret.core.atlas as atlas
    from sklearn import mixture
    from sciplotlib import style as spstyle


def plot_umap(umap_coord, label, color, color_name, color_map='plasma',label_order=None):
    """Plots the UMAP figures for traces grouped by label values.

    This function generates the UMAP figures used in the META-SiM paper.

    Args:
        umap_coord: The 2D coordinates of the umap, as a numpy tensor.
        label: The list of label values, as a numpy array.
        color: The color label for all traces, as a numpy array.
        color_name: The name of the color label, such as entropy.
        color_map: The name of colormap. If not set, use plasma as default.
        label_order: A list of labels to plot. If set, order the figures with these label values.

    Returns:
        A figure of UMAP plots.
    """
    if label_order is None:
        label_order = sorted(list(set(label.tolist())))
    counter = 0
    ax_min = np.min(umap_coord)
    ax_max = np.max(umap_coord)
    with plt.style.context(spstyle.get_style('nature-reviews')):
        fig, axes = plt.subplots(ncols=len(label_order), nrows=1, figsize=(5 * len(label_order), 5));
    for l in label_order:
        idx = (label == l)
        z = umap_coord[idx, ...]
        with plt.style.context(spstyle.get_style('nature-reviews')):
            if not isinstance(axes, np.ndarray):
                axes = [axes]  # handles single figure case.
            ax = axes[counter]
            counter += 1
            s = ax.scatter(z[:, 0], z[:, 1], c=color[idx], s=30, alpha=0.6, linewidths=0, cmap=color_map, vmin=np.quantile(color, 0.01), vmax=np.quantile(color, 0.99))
            ax.set_xlim([ax_min, ax_max])
            ax.set_ylim([ax_min, ax_max])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_aspect('equal')
            ax.set_title(f'{l}')
            ax.text(x=ax.get_xlim()[0]+ 0.1, y=ax.get_ylim()[0] + 0.1, s=f'n={z.shape[0]}', fontdict={'size': 12})

    color_bar = fig.colorbar(s, ax=axes, label=color_name)
    color_bar.set_alpha(1)
    color_bar.draw_all()
    return fig, axes


def get_umap_reducer(embedding, n_neighbors=30, n_epochs=500):
    """Fits a 2-D UMAP reducer given embeddings.

    In this function, we fix several parameters in the UMAP model
    in order to obtain stable and robust UMAP output.
    The fixed parameters are:
        1. random_state
        2. negative_sample_rate
        3. min_dist

    Args:
        embedding: The embedding tensor.
        n_neighbors: The number of neighbors used for UMAP optimization.
        n_epochs: The epochs number of UMAP optimization.

    Returns:
        A umap reducer. The reducer can be used by:
        umap_coord = reducer.transform(embedding)
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        reducer = umap.UMAP(
            n_epochs=n_epochs, n_neighbors=n_neighbors, min_dist=0.0,
            random_state=np.random.RandomState(60), n_components=2,
            negative_sample_rate=15)
        reducer.fit(embedding)
    return reducer


def plot_atlas_reference():
    """Plots the Atlas boundaries and legends in a reference figure.

    This figure can be used as a reference when reading the Atlas UMAP.

    Returns:
        A figure of the Atlas reference.
    """
    return _plot_atlas_reference()

def _plot_atlas_reference(
        fig=None, ax=None, clean_only=False, use_legend=True,
):
    """Internal function for ploting Atlas boundaries and legends in a reference figure.

    This figure can be used as a reference when reading the Atlas UMAP.

    Returns:
        A figure of the Atlas reference.
    """
    saved_atlas_density_models = atlas.get_atlas_density_model()
    with plt.style.context(spstyle.get_style('nature-reviews')):
        if fig is None:
            fig = plt.figure(figsize=(9, 6))
        if ax is None:
            ax = fig.add_axes([0.2, 0.17, 0.68, 0.7])
        contour_counter = 0
        for l in saved_atlas_density_models:
            xmin = -10
            xmax = 22
            ymin = -11
            ymax = 18
            color = plt.cm.tab20.colors + plt.cm.tab20b.colors[::2]
            x = np.linspace(xmin, xmax, num=200)
            y = np.linspace(ymin, ymax, num=200)
            X, Y = np.meshgrid(x, y)
            clf = saved_atlas_density_models[l]
            XX = np.array([X.ravel(), Y.ravel()]).T
            Z = clf.score_samples(XX)
            Z = np.exp(Z.reshape(X.shape))
            cmap = plt_colors.LinearSegmentedColormap.from_list("", [color[contour_counter], color[contour_counter], color[contour_counter]])
            if (not clean_only) or ('-c-' in l):
                CS = ax.contour(
                    X, Y, Z, cmap=cmap, levels=[1e-2]
                )
                if use_legend and CS.collections:
                    CS.collections[0].set_label(l)
            contour_counter += 1
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.set_aspect('equal')
        ax.grid(True)
        if use_legend:
            ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), markerscale=4)
            plt.title('smFRET Atlas')


def plot_atlas(embedding, label, color, color_name, label_order=None):
    """Plots the Atlas UMAP figures for traces grouped by label values.

    This function generates the Atlas figures used in the META-SiM paper.

    Args:
        embedding: The embedding of traces, generated by model.encode().
        label: The list of label values, as a numpy array.
        color: The color label for all traces, as a numpy array.
        color_name: The name of the color label, such as entropy.
        label_order: A list of labels to plot. If set, order the figures with these label values.

    Returns:
        A figure of Atlas UMAP plots.
    """
    umap_coord = atlas.get_atlas_2d(embedding)
    fig, axes = plot_umap(umap_coord, label=label, color=color,
                          color_name=color_name, label_order=label_order)
    saved_atlas_density_models = atlas.get_atlas_density_model()
    for ax in axes:
        _plot_atlas_reference(
            fig=fig, ax=ax, clean_only=True, use_legend=False,
        )
    return fig, axes


def plot_fret_histograms(fret_efficiency, label, label_order=None):
    """Plots the FRET histogram figures for traces grouped by label values.

    This function generates the Atlas figures used in the META-SiM paper.

    Args:
        fret_efficiency: The list of FRET efficiency, generated by get_fret_efficiency().
        label: The list of label values, as a numpy array.
        label_order: A list of labels to plot. If set, order the figures with these label values.

    Returns:
        A figure of FRET efficiency plots.
    """
    if label_order is None:
        label_order = sorted(list(set(label.tolist())))

    with plt.style.context(spstyle.get_style('nature-reviews')):
        fig, axes = plt.subplots(ncols=len(label_order), nrows=1, figsize=(5 * len(label_order), 5));
        if not isinstance(axes, np.ndarray):
            axes = [axes]

    count = 0
    for l in label_order:
        idx = (label == l)
        values = []
        for i, found in enumerate(idx):
            if found:
                value = fret_efficiency[i]
                if value.shape[0] > 0:
                    values.append(value)
        if values:
            values = np.concatenate(values, axis=0)
        with plt.style.context(spstyle.get_style('nature-reviews')):
            ax = axes[count]
            ax.hist(values, bins=50, range=[-0.3, 1.2], color=(85 / 255, 154 / 255, 209 / 255), alpha=1, edgecolor='w');
            ax.set_xlim([-0.3, 1.2])
            ax.set_title(l)
            ax.set_xlabel('FRET Value')
            ax.set_ylabel('Count')
            ax.text(x=ax.get_xlim()[0]+ 0.1, y=ax.get_ylim()[0] + 0.1, s=f'n={np.sum(idx)}', fontdict={'size': 12})
            count += 1
    return fig, axes






"""Library for linear algebra operations in META-SiM."""

import numpy as np
import tensorflow as tf
from dataclasses import dataclass, field
from typing import List, Tuple, Dict, Union, Optional


def principal_project(embedding, kernel, effective_rank_fraction=0.25):
    """Projects the embeddings into the principal directions of kernel.

    This is the main function for principal projection. The kernel can
    be from a single downtream task or the concatenated kernel across
    multiple tasks.

    Args:
        embedding: The original embedding vectors as a numpy tensor.
        kernel: A numpy tensor of kernel.
        effective_rank_fraction: The fraction of 2nd order momentum to keep for low rank approximation.

    Returns:
        The projected embeddings.
    """


    u1, s1, _ = np.linalg.svd(kernel, full_matrices=False, hermitian=False)

    effective_rank_1 = u1.shape[1]
    for i in range(s1.shape[0]):
        if np.sum(s1[i:] ** 2) < (1.0 - effective_rank_fraction) * np.sum(s1 ** 2):
            effective_rank_1 = i
            break
    return np.einsum('ij,jk->ik', embedding, u1[:, :effective_rank_1])


def get_kernel_alignment(kernel_1, kernel_2, effective_rank_fraction=0.25):
    """Measures the alignment between two linear kernels.

    This function is used to measure the alignment between two tasks.

    Args:
        kernel_1: A numpy tensor of kernel 1.
        kernel_2: A numpy tensor of kernel 2.
        effective_rank_fraction: The fraction of 2nd order momentum to keep for low rank approximation.

    Returns:
        The alignment coefficient between [0, 1].
    """
    u1, s1, _ = np.linalg.svd(kernel_1, full_matrices=False, hermitian=False)
    u2, s2, _ = np.linalg.svd(kernel_2, full_matrices=False, hermitian=False)

    effective_rank_1 = u1.shape[1]
    for i in range(s1.shape[0]):
        if np.sum(s1[i:] ** 2) < (1.0 - effective_rank_fraction) * np.sum(s1 ** 2):
            effective_rank_1 = i
            break
    effective_rank_2 = u2.shape[1]
    for i in range(s2.shape[0]):
        if np.sum(s2[i:] ** 2) < (1.0 - effective_rank_fraction) * np.sum(s2 ** 2):
            effective_rank_2 = i
            break
    coeff = np.matmul(np.transpose(u1[:, :effective_rank_1]), u2[:, :effective_rank_2])
    return np.linalg.norm(coeff, 2)



"""Library for downstream tasks using META-SiM."""

import os
import json
import numpy as np
from dataclasses import dataclass, field
from typing import List, Tuple, Dict, Union, Optional, Any
from pathlib import Path
from sklearn.linear_model import LogisticRegression

VERSION = 'METASIM-1.0.0'  # First version of META-SiM

 # Group training tasks according to area of FRET feature.
TASKS = [
     ['DWELLTIME', 'TRANSITIONPROBQUICK', 'NBDQUICK', 'NBD', 'TRANSITIONRATE'],
     ['FRETSTATECOUNT', ],
     ['FRETSTATEHIGHEST', 'FRETSTATELOWEST'],
     ['MULTISTEP', 'COINCIDENCE',],
     ['ACCEPTORLIFETIME'],
     ['DONORLIFETIME'],
     ['NOISELEVEL', 'BACKGROUND'],
     ['SINGLECHANNELDWELLTIME', 'SINGLECHANNELLIFETIME',],
     ['SINGLECHANNELPHOTOBLEACHINGSTEPSTRACE',],
     ['SINGLECHANNELNOISELEVEL',],
]

NAMES = [
     'Kinetic\nRate',
     'FRET State\nNumber',
     'FRET\nValue',
     'Photobleaching\nSteps',
     'Acceptor\nLifetime',
     'Donor\nLifetime',
     'Noise',
     'Single Ch.\nKinetic Rate',
     'Single Ch.\nPhotobleaching',
     'Single Ch.\nNoise',
 ]

@dataclass
class Weights:
    kernel: List[List[float]]
    bias: List[float]

    def to_dict(self):
        return {"kernel": self.kernel, "bias": self.bias}

    @classmethod
    def from_dict(cls, data):
        if isinstance(data["kernel"], np.ndarray):
            data["kernel"] = data["kernel"].tolist()
        if isinstance(data["bias"], np.ndarray):
            data["bias"] = data["bias"].tolist()
        return cls(kernel=data["kernel"], bias=data["bias"])


@dataclass
class TuningSetting:
    id: int
    optimizer: str
    learningRate: float
    batchSize: int
    epochs: int
    validationSplit: float
    l2Regularization: float

    def to_dict(self):
        return {
            "id": self.id,
            "optimizer": self.optimizer,
            "learningRate": self.learningRate,
            "batchSize": self.batchSize,
            "epochs": self.epochs,
            "validationSplit": self.validationSplit,
            "l2Regularization": self.l2Regularization,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            id=data["id"],
            optimizer=data["optimizer"],
            learningRate=data["learningRate"],
            batchSize=data["batchSize"],
            epochs=data["epochs"],
            validationSplit=data["validationSplit"],
            l2Regularization=data["l2Regularization"],
        )


@dataclass
class ModelConfig:
    weights: Weights
    lookupList: List[Any]
    version: Optional[str] = VERSION
    tuningSetting: Optional[TuningSetting] = None

    def to_dict(self):
        if isinstance(self.lookupList, np.ndarray):
            self.lookupList = self.lookupList.tolist()
        for i, e in enumerate(self.lookupList):
            if getattr(e, 'dtype', None) == np.int64:
                self.lookupList[i] = int(self.lookupList[i])
        data = {
            "weights": self.weights.to_dict(),
            "lookupList": self.lookupList,
            "version": self.version,
        }
        if self.tuningSetting:
            data["tuningSetting"] = self.tuningSetting.to_dict()
        return data

    @classmethod
    def from_dict(cls, data):
        weights = Weights.from_dict(data["weights"])
        tuning_setting = TuningSetting.from_dict(data["tuningSetting"]) if "tuningSetting" in data else None
        return cls(weights=weights, lookupList=data["lookupList"], version=data["version"], tuningSetting=tuning_setting)

    def save_to_json(self, file_path: Path):
        with open(file_path, "w") as f:
            json.dump(self.to_dict(), f, indent=4) # indent makes it readable

    @classmethod
    def load_from_json(cls, file_path: Path):
        with open(file_path, "r") as f:
            data = json.load(f)
        return cls.from_dict(data)

    @classmethod
    def load_from_task(cls, task: str):
        """Loads pre-trained downstream tasks."""
        file_path = os.path.join(
            os.path.dirname(__file__),
            'downstream_task_model_configs',
            f'{task}.json',
        )
        return cls.load_from_json(file_path)

    def predict(self, embedding):
        """Generates predicted labels."""
        logits = np.matmul(
            embedding, self.weights.kernel) + self.weights.bias
        idx = np.argmax(logits, -1)
        pred = np.array(self.lookupList)[idx]
        return pred

    def predict_proba(self, embedding):
        """Generates predicted probability distribution."""
        logits = np.matmul(
            embedding, self.weights.kernel) + self.weights.bias
        pred = softmax(logits)
        return pred

    def predict_regression(self, embedding, optimize_for='RMSE'):
        """Generates regression prediction for continuous labels.

        The output is decided by the final target.
        - If optimize for MAE, predict the median of predicted distribution
        - If optimize for RMSE, predict the expectation of distribution

        Args:
            embedding: The embedding numpy tensor.
            optimize_for: 'RMSE' or 'MAE'.
        """
        if optimize_for.upper() not in ['MAE', 'RMSE']:
            raise ValueError(
                f'Regression can only optimize for MAE or RMSE, got {optimize_for}',
            )
        if optimize_for.upper() == 'RMSE':
            pred = np.einsum(
                'ij,j->i',
                self.predict_proba(embedding),
                np.array(self.lookupList),
            )
            return pred
        elif optimize_for.upper() == 'MAE':
            cum_prob = np.cumsum(self.predict_proba(embedding), axis=-1)
            idx = np.argmax(cum_prob >= 0.5, axis=-1)
            return np.array(self.lookupList)[idx]


def softmax(x):
    """
    Numerically stable softmax function for a 2D array.

    Args:
        x: A 2D numpy array.

    Returns:
        A 2D numpy array with the softmax applied along the last axis.
    """

    # Find the maximum value along the last axis (axis=1) for each row.  This is crucial for numerical stability.
    max_x = np.max(x, axis=1, keepdims=True)  # keepdims=True is important for proper broadcasting

    # Subtract the maximum value from x before exponentiating. This prevents potential overflow issues.
    shifted_x = x - max_x

    # Exponentiate the shifted values.
    exp_x = np.exp(shifted_x)

    # Calculate the sum of exponentials along the last axis.
    sum_exp_x = np.sum(exp_x, axis=1, keepdims=True)

    # Normalize by dividing by the sum of exponentials.
    softmax_x = exp_x / sum_exp_x

    return softmax_x


def train_classification(
        embedding, label, class_weight=None, max_iter=5000,
):
    """Trains a classification head for downstream task.

    This function uses logistic regression to train a head for trace
    level classification task.

    Args:
        embedding: The numpy tensor of trace embeddings.
        label: The numpy array of trace labels.
        class_weight: An optional dict to contain weights for each class.
        max_iter: An optional integer to define max iterations for optimization.

    Returns:
        A ModelConfig object for the trained task head.
    """
    clf = LogisticRegression(
        max_iter=max_iter, tol=1e-6, class_weight=class_weight,
    ).fit(embedding, label)
    if clf.coef_.shape[0] == 1:
        clf.coef_ = np.concatenate([np.zeros_like(clf.coef_), clf.       coef_], axis=0)
        clf.intercept_ = np.concatenate([np.zeros_like(clf.intercept_),  clf.intercept_], axis=0)
    num_classes = clf.coef_.shape[0]
    weights = Weights(
        kernel=clf.coef_.T.tolist(),
        bias=clf.intercept_.tolist(),
    )
    model = ModelConfig(
        weights=weights,
        lookupList=clf.classes_.tolist(),
    )
    return model


def train_regression(
        embedding, label, max_iter=5000, num_bins=8,
):
    """Trains a regression model.

    This function trains a regression model. In META-SiM, this is done via
    quantization, i.e. bucketing labels into classes.

    The output is decided by the final target. We use:
        - If optimize for MAE, the median of predicted distribution
        - If optimize for RMSE, the expectation of predicted distribution

    Args:
        embedding: The embedding numpy tensor.
        label: The numpy array of label.
        max_iter: The max iterations of optimization.
        num_bins: The number of bins for quantization.

    Return:
        A nmpy array of prediction.
    """
    lower_limit = np.min(label)
    upper_limit = np.max(label)
    quanta = (upper_limit - lower_limit) / num_bins
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    quantized_label = np.digitize(label, quantize_bins, right=True)
    model = train_classification(
        embedding, quantized_label, max_iter=max_iter,
    )
    model.lookupList = (quantize_bins + 0.5 * quanta).tolist()
    model.lookupList[-1] -= 0.5 * quanta  # caps the max
    model.lookupList = [lower_limit] + model.lookupList  # right is (a, b].
    return model


def moving_majority_vote(arr, window_size):
    """
    Applies a moving majority vote filter to a 1D numpy array.

    Args:
        arr: The input 1D numpy array.
        window_size: The size of the moving window. Must be an odd integer.

    Returns:
        A numpy array of the same size as the input, containing the majority vote
        at each position.  Returns the original array if window_size is not a
        positive odd integer or if window_size is greater than the array size.
    """

    if not isinstance(window_size, int) or window_size <= 0 or window_size % 2 == 0 or window_size > len(arr):
      return arr  # Return original array if window size is invalid

    n = len(arr)
    result = np.zeros(n, dtype=arr.dtype)  # Ensure output has the same data type

    half_window = window_size // 2

    for i in range(n):
        window_start = max(0, i - half_window)
        window_end = min(n, i + half_window + 1)  # +1 because range is exclusive of end

        window = arr[window_start:window_end]

        if len(window) == 0:  # Handle edge cases where the window is empty.
            result[i] = arr[i] # or some default value.
            continue


        counts = np.bincount(window) # efficient counting of occurrences

        if len(counts) == 0: # Handle edge cases where the window is empty.
            result[i] = arr[i] # or some default value.
            continue

        majority_index = np.argmax(counts) # index of the most frequent value
        result[i] = majority_index

    return result



"""Analyze a single smFRET trace."""

import logging
import numpy as np
import dataclasses
import scipy.signal


MEDIAN_FILTER_WINDOW = 5  # Window size for smoothing traces with a median filter
NUM_STD_LIFETIME = 5  # Threshold for calculating lifetime
NUM_STD_MULTI_STEP = 2  # Threshold for detecting multi-step photobleaching
NUM_STD_BLINK = 1  # Threshold for detecting Cy3 blinking
MIN_FRET = 0.1  # Min FRET value to be considered valid FRET region
MIN_FRET_LENGTH = 5  # Min number of continuous frames to be considered valid FRET region
EPS = np.finfo(np.float32).eps  # Numeric tolerance for float32


@dataclasses.dataclass
class TraceStatistics:
    mean_total_intensity: float = None
    max_fret: float = None  # Highest FRET value
    first_fret: float = None  # FRET at first frame
    fret_events: int = None  # Number of FRET events
    acceptor_lifetime: float = None  # FRET acceptor lifetime
    donor_lifetime: float = None  # FRETdonor lifetime
    trace_lifetime: float = None  # FRET trace lifetime, i.e. end of the trace.
    correlation: float = None  # Correlation of Fluorophore
    correlation_derivative: float = None  # Correlation of Fluorophore derivative
    snr_background: float = None  # Signal-to-noise for the background
    snr_signal: float = None  # Signal-to-noise for the signal
    nnr_ratio: float = None  # # signal noise to background noise ratio
    background_noise: float = None  # Background noise
    cy3_blinks:  int = None  # Number of Cy3 blinks
    multi_step_photobleaching: bool = None  # Whether exists multi step photobleaching
    mean_acceptor: float = None  # Mean acceptor intensity
    mean_donor: float = None  # Mean donor intensity
    mean_fret: float = None  # Mean FRET value
    
    def asdict(self):
        """Converts the dataclass into a Python dictionary."""
        return dataclasses.asdict(self)
 
    def astuple(self):
        """Converts the dataclass into a Python tuple.
        
        The statistics names will be lost by this conversion.
        """
        return dataclasses.astuple(self)
    
    def __array__(self):
        return np.array(self.astuple())

    def __len__(self):
        return dataclasses.astuple(self).__len__()

    def __getitem__(self, item):
        return dataclasses.astuple(self).__getitem__(item)


def analyze_trace(trace) -> TraceStatistics:
    """Analyze a smFRET trace by generating the statistics of its signal.""" 
    trace_lifetime = calculate_trace_lifetime(trace)
    if trace_lifetime is None:
        return TraceStatistics()  # All the rest calculations are unsafe. Discard them now.
    multi_step_photobleaching = detect_multi_step_photobleaching(trace, trace_lifetime)
    background_noise = calculate_background_noise(trace, trace_lifetime)
    cy3_blinks = count_cy3_blinks(trace, trace_lifetime)
    donor_lifetime = calculate_donor_lifetime(trace, trace_lifetime)
    acceptor_lifetime = calculate_acceptor_lifetime(trace, trace_lifetime)
    mean_donor, mean_acceptor, mean_total_intensity = calculate_mean_intensities(trace, trace_lifetime)
    snr_background, snr_signal, nnr_ratio = calculate_snrs(trace, trace_lifetime)
    correlation, correlation_derivative = calculate_correlations(trace, trace_lifetime)
    mean_fret, max_fret, first_fret = calculate_fret_statistics(trace, trace_lifetime)
    fret_events = count_fret_events(trace, trace_lifetime)
    trace_statistics = TraceStatistics(
        trace_lifetime=trace_lifetime,
        mean_total_intensity=mean_total_intensity,
        max_fret=max_fret,
        first_fret=first_fret,
        fret_events=fret_events,
        acceptor_lifetime=acceptor_lifetime,
        donor_lifetime=donor_lifetime,
        correlation=correlation,
        correlation_derivative=correlation_derivative,
        snr_background=snr_background,
        snr_signal=snr_signal,
        nnr_ratio=nnr_ratio,
        background_noise=background_noise,
        cy3_blinks=cy3_blinks,
        multi_step_photobleaching=multi_step_photobleaching,
        mean_acceptor=mean_acceptor,
        mean_donor=mean_donor,
        mean_fret=mean_fret,
    )
    return trace_statistics
    
    
def calculate_trace_lifetime(trace):
    """Calculates donor lifetime for a smFREt trace."""
    smooth_total = scipy.signal.medfilt(trace.total, MEDIAN_FILTER_WINDOW)
    gradient_total = np.gradient(smooth_total)  # 2nd order discrete differenctiation
    mean_gradient_total = np.mean(gradient_total)
    std_gradient_total = np.std(gradient_total)
    
    non_outliers = abs(gradient_total - mean_gradient_total) <= 6.0 * std_gradient_total 
    threshold = mean_gradient_total - NUM_STD_LIFETIME * np.std(gradient_total[non_outliers])
    lifetime = np.argwhere(gradient_total <= threshold)
    if lifetime.size > 0:
        return max(2, lifetime[-1, 0]) # Found the last drop in the total intensity.
    else:
        return None

def detect_multi_step_photobleaching(trace, trace_lifetime):
    """Detect the frame of multi-step photobleaching events in a trace."""
    smooth_total = scipy.signal.medfilt(trace.total, MEDIAN_FILTER_WINDOW)
    gradient_total = np.gradient(smooth_total)  # 2nd order discrete differenctiation
    mean_gradient_total = np.mean(gradient_total)
    std_gradient_total = np.std(gradient_total)
    
    
    threshold = mean_gradient_total - NUM_STD_MULTI_STEP * std_gradient_total
    # Ignores events right at the begining or after donor lifetime.
    events = np.argwhere(gradient_total[5 : trace_lifetime] <= threshold) + 1
    
    for frame in events[::-1]:
        if frame[0] - 1 > 2:
            min_total_before = np.min(smooth_total[2 : frame[0] - 1])
        else:
            continue
        if frame[0] + 1 < trace_lifetime:
            max_total_after = np.max(smooth_total[frame[0] + 1: trace_lifetime])
        else:
            continue
        if min_total_before >= max_total_after:
            return True  # detected at least one multi-step photobleach
    return False 

def calculate_background_noise(trace, trace_lifetime):
    """Calculates the background noise of a smFRET trace."""
    if (trace_lifetime + 10) < trace.donor.size:
        start = trace_lifetime + 5  # Safely after the donor lifetime.
        noise = np.std(trace.donor[start:]) + np.std(trace.acceptor[start:])
        return noise
    else:
        return None

def count_cy3_blinks(trace, trace_lifetime: int):
    """Counts the number of Cy3 blinks in a smFRET trace."""
    if (trace_lifetime + 5) >= trace.donor.size:
        return None
    std_background = np.std(trace.total[trace_lifetime + 5:])
    blinks = trace.total[1:trace_lifetime - 1] <= NUM_STD_BLINK * std_background
    blink_starts = np.bitwise_and(blinks[:-1], ~blinks[1:])  # finds the start of each blink
    return np.sum(blink_starts)

def calculate_donor_lifetime(trace, trace_lifetime: int):
    """Calculates the length of the "donor-alive" region."""
    if (trace_lifetime + 5) >= trace.donor.size:
        return None
    std_background = np.std(trace.total[trace_lifetime + 5:])
    donor_alive = trace.total[:trace_lifetime - 1] > NUM_STD_BLINK * std_background
    return np.sum(donor_alive) 
    
def calculate_mean_intensities(trace, trace_lifetime: int):
    """Calculates the mean donor/acceptor/total intensities."""
    if (trace_lifetime + 5) >= trace.donor.size:
        return None, None, None
    std_background = np.std(trace.total[trace_lifetime + 5:])
    donor_alive = trace.total[:trace_lifetime] > NUM_STD_BLINK * std_background
    donor_alive_no_falling_edges = donor_alive[:-2] & donor_alive[1:-1] & donor_alive[2:]
    mean_donor = np.mean(trace.donor[:trace_lifetime - 2][donor_alive_no_falling_edges])
    mean_acceptor = np.mean(trace.acceptor[:trace_lifetime - 2][donor_alive_no_falling_edges])
    mean_total = mean_donor + mean_acceptor
    return (mean_donor, mean_acceptor, mean_total)

def calculate_snrs(trace, trace_lifetime: int):
    """Calculates the signal-to-noise ratios of a smFRET trace."""
    if (trace_lifetime + 5) >= trace.donor.size:
        return None, None, None
    std_background = np.std(trace.total[trace_lifetime + 5:])
    donor_alive = trace.total[:trace_lifetime] > NUM_STD_BLINK * std_background
    donor_alive_no_falling_edges = donor_alive[:-2] & donor_alive[1:-1] & donor_alive[2:]
    donor = trace.donor[:trace_lifetime - 2][donor_alive_no_falling_edges]
    acceptor = trace.acceptor[:trace_lifetime - 2][donor_alive_no_falling_edges]
    mean_total = np.mean(donor + acceptor)
    snr = mean_total / std_background  # signal to background noise
    snr_signal = mean_total / np.std(donor + acceptor)  # signal to signal noise
    nnr = np.std(donor + acceptor) / std_background  # signal noise to background noise ratio
    return (snr, snr_signal, nnr)
    
def calculate_correlations(trace, trace_lifetime: int):
    """Calculates the correlations of donor v.s. acceptor, and between their derivatives."""
    if (trace_lifetime + 5) >= trace.donor.size:
        return None, None
    std_background = np.std(trace.total[trace_lifetime + 5:])
    donor_alive = trace.total[:trace_lifetime] > NUM_STD_BLINK * std_background
    donor_alive_no_falling_edges = donor_alive[:-2] & donor_alive[1:-1] & donor_alive[2:]
    if np.sum(donor_alive_no_falling_edges) < MIN_FRET_LENGTH:
        return (None, None)  # Length of donor-alive region too small to be meaningful
    donor = trace.donor[:trace_lifetime - 2][donor_alive_no_falling_edges]
    acceptor = trace.acceptor[:trace_lifetime - 2][donor_alive_no_falling_edges]
    gradient_donor = np.gradient(donor)
    gradient_acceptor = np.gradient(acceptor)
    correlation = np.corrcoef(donor, acceptor)[0, 1]
    gradient_correlation = np.corrcoef(gradient_donor, gradient_acceptor)[0, 1]
    return (correlation, gradient_correlation)


def run_length_encode(v: np.ndarray):
    """Generates the run-length encoding for a 1-D numpy array."""
    n = v.size
    encoding = np.zeros(v.size)
    running_sum = 0
    for i in range(n)[::-1]:
        if v[i] == 0:
            running_sum = 0
        else:
            running_sum += 1
            encoding[i] = running_sum
    return encoding

def run_length_filter(v: np.ndarray, min_length: int):
    """Filters a 1-D numpy array by minimum run length of a continuous region."""
    encoding = run_length_encode(v)
    filtered_v = np.zeros_like(v)
    running_max = 0
    for i in range(v.size):
        if v[i] == 0:
            running_max = 0
        else:
            running_max = max(running_max, encoding[i])
            filtered_v[i] = (running_max >= min_length)
    return filtered_v
 
def calculate_fret_statistics(trace, trace_lifetime: int):
    """Calculates the average/max/first-frame FRET value."""
    valid_fret_bool = trace.fret[:trace_lifetime] >= MIN_FRET
    valid_fret_bool = run_length_filter(valid_fret_bool, min_length=MIN_FRET_LENGTH)
    valid_fret = trace.fret[:trace_lifetime][valid_fret_bool]
    if valid_fret.size > 0:
        mean_fret = np.mean(valid_fret)
        max_fret = np.max(valid_fret)
        first_fret = valid_fret[0]
        return (mean_fret, max_fret, first_fret)
    else:
        return (None, None, None)

def calculate_acceptor_lifetime(trace, trace_lifetime: int):
    """Calculates the acceptor lifetime of a smFRET trace."""
    valid_fret_bool = trace.fret[:trace_lifetime] >= MIN_FRET
    valid_fret_bool = run_length_filter(valid_fret_bool, min_length=MIN_FRET_LENGTH)
    acceptor_lifetime = np.sum(valid_fret_bool)
    return acceptor_lifetime

def count_fret_events(trace, trace_lifetime: int):
    """Counts the number of FRET events in a smFRET trace."""
    encoding = run_length_encode(trace.fret[:trace_lifetime] >= MIN_FRET)
    n_fret = np.sum(encoding == 1)  # Counts the end of each FRET
    return n_fret
 


"""Library for META-SiM models."""
import os
import tensorflow as tf
from tqdm.auto import tqdm
from metasim.fret.core.legacy.tf_layers import Attention
from metasim.fret.core.legacy.tf_layers import Conv
from metasim.fret.core.legacy.tf_layers import Summary
from metasim.fret.core.legacy.tf_layers import PrependTaskToken
from metasim.fret.core.legacy.tf_layers import Embedding
from metasim.fret.core.legacy.tf_layers import PositionEmbedding

# import Keras 2.0
try:
    import tf_keras as keras  # backward compatible with Keras 3.0
except:
    # If tf_keras is not available, use old syntax to import
    import tensorflow.keras as keras

CHECKPOINT_PATH = 'checkpoints/encoder-20240111-045226.h5'
CHECKPOINT_PATH = os.path.join(os.path.dirname(__file__), CHECKPOINT_PATH)
BATCH_SIZE = 16
WINDOW_SIZE = 100


class Model:
    """Class for META-SiM model."""
    def __init__(self, checkpoint_path=CHECKPOINT_PATH):
        keras.mixed_precision.set_global_policy('mixed_bfloat16')
        self.encoder = keras.models.load_model(
            checkpoint_path,
            compile=False,
            custom_objects={'PositionEmbedding': PositionEmbedding},
        )

    def _get_stacked_tensors(self, tensors, trim=False):
        """Groups tensors according to their trace's length.

        This function is used to optimize the inference efficienty. Traces of the
        same length can be stored in one tensor for faster inference.
        """
        batches = []
        for tensor in tensors:
            if trim and tensor.shape[0] >= WINDOW_SIZE:
                trim_size = tensor.shape[0] // WINDOW_SIZE * WINDOW_SIZE
                tensor = tensor[:trim_size, ...]
            if not batches:
                batches.append([tensor])
                continue
            if batches[-1][-1].shape[0] == tensor.shape[0]:
                batches[-1].append(tensor)
            else:
                batches.append([tensor])
        # stack each batch
        stacked_batches = [tf.stack(batch, axis=0) for batch in batches]
        return stacked_batches

    def encode(self, dataset):
        """Encodes a dataset into a numpy tensor of trace level embeddings.

        This is the main method to use META-SiM model.
        """
        if hasattr(self.encoder.layers[-1], 'framewise'):
            # This is a converted encoder
            self.encoder.layers[-1].framewise = False
        tensors = dataset.to_tensors()
        stacked_batches = self._get_stacked_tensors(tensors, trim=True)
        embeddings = []
        with tqdm(total=len(dataset.traces)) as pbar:
            pbar.set_description("Encoding traces in dataset %s" % dataset.title)
            for stacked_batch in stacked_batches:
                embeddings.extend(tf.unstack(
                    self.encoder.predict(
                        stacked_batch, batch_size=BATCH_SIZE, verbose=0,
                    )
                ))
                pbar.update(stacked_batch.shape[0])
        return tf.stack(embeddings, axis=0).numpy()

    def __call__(self, dataset):
        """Alias for encode."""
        return self.encode(dataset)

    def encode_frames(self, dataset):
        """Encodes traces into a list of frame-level embeddings.

        This method is useful when you want to build frame-level downstream models.
        """
        self.encoder.layers[-1].framewise = True
        tensors = dataset.to_tensors()
        stacked_batches = self._get_stacked_tensors(tensors)
        embeddings = []
        with tqdm(total=len(dataset.traces)) as pbar:
            pbar.set_description("Encoding traces in dataset %s" % dataset.title)
            for stacked_batch in stacked_batches:
                embeddings.extend(tf.unstack(
                    self.encoder.predict(
                        stacked_batch, batch_size=BATCH_SIZE, verbose=0,
                    )
                ))
                pbar.update(stacked_batch.shape[0])

        # Trims the embeddings' paddings
        # Embeddings can be longer than the trace length due to zero padding of tokenizer.
        for i, tensor in enumerate(tensors):
            embeddings[i] = embeddings[i][:tensor.shape[0], ...].numpy()
        return embeddings




"""Classes and functions for learning and HP tuning with TensorFlow."""

import tensorflow as tf


class LRSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):

    def __init__(self, max_learning_rate, warmup_steps, decay_steps):
        self.max_learning_rate = tf.cast(max_learning_rate, tf.float32)
        self.warmup_steps = tf.cast(warmup_steps, tf.float32)
        self.decay_steps = tf.cast(decay_steps, tf.float32)

    def __call__(self, step):
        step = tf.cast(step, tf.float32)
        if step <= self.warmup_steps:
            return self.max_learning_rate / self.warmup_steps * step
        elif step <= self.warmup_steps + self.decay_steps:
            return (self.max_learning_rate / self.decay_steps 
                    * (self.decay_steps + self.warmup_steps - step))
        else:
            return 0.0
    
    def get_config(self):
        config = {
            "max_learning_rate": self.max_learning_rate,
            "warmup_steps": self.warmup_steps,
            "decay_steps": self.decay_steps,
        }
        return config


"""Library for data structures in META-SiM."""
from openfret import Trace
from openfret import Dataset
import tensorflow as tf

TIME_DENOMINATOR = 2000   # A constant rescaler used in training


def _trace_data_to_tensor(donor, acceptor):
    """Helper function to convert data into tensor."""
    if acceptor is None:
        raise ValueError('Acceptor cannot be set to None.')
    acceptor = tf.cast(acceptor, tf.float32)
    if donor is None:
        # Fill in zeros for single color scenario.
        # In training, acceptor is used to host the single channel data.
        donor = tf.zeros_like(acceptor)
    else:
        donor = tf.cast(donor, tf.float32)
    total = donor + acceptor
    tensor = tf.stack([donor, acceptor, total], axis=-1)
    total_max = tf.reduce_max(total, axis=(-1))
    total_max = tf.expand_dims(total_max, axis=-1)
    tensor = tensor / total_max
    time = tf.expand_dims((tf.range(tensor.shape[0], dtype=tf.float32) + 1.0) / TIME_DENOMINATOR, axis=-1)
    tensor = tf.concat([tensor, time], axis=-1)
    return tensor


class OneColorDataset(Dataset):
    """Class for single color dataset."""
    def __init__(self, dataset, channel_type):
        self.__dict__.update(dataset.__dict__)  # copies from parent class
        self.channel_type = channel_type

    def trace_to_tensor(self, trace: Trace):
        """Converts a trace into a TensorFlow tensor."""
        signal = None
        for channel in trace.channels:
            if channel.channel_type == self.channel_type:
                signal = channel.data
        if signal is None:
            raise ValueError(f"{self.channel_type} channel must exist in all traces.")
        return _trace_data_to_tensor(donor=None, acceptor=signal)

    def to_tensors(self):
        """Converts a dataset into a list of tensors."""
        tensor_list = []
        for trace in self.traces:
            with tf.device('/CPU:0'):   # reduces communication overhead.
                tensor_list.append(self.trace_to_tensor(trace))
        return tensor_list


class TwoColorDataset(Dataset):
    """Class for two color FRET dataset."""
    def __init__(self, dataset, donor_channel_type="donor", acceptor_channel_type="acceptor"):
        self.__dict__.update(dataset.__dict__)  # copies from parent class
        self.donor_channel_type = donor_channel_type
        self.acceptor_channel_type = acceptor_channel_type

    def trace_to_tensor(self, trace: Trace):
        """Converts a trace into a TensorFlow tensor."""
        donor = None
        acceptor = None
        for channel in trace.channels:
            if channel.channel_type == self.donor_channel_type:
                donor = channel.data
            if channel.channel_type == self.acceptor_channel_type:
                acceptor = channel.data
        if donor is None:
            raise ValueError(f"{self.donor_channel_type} channel must exist in all traces.")
        if acceptor is None:
            raise ValueError(f"{self.acceptor_channel_type} channel must exist in all traces.")
        return _trace_data_to_tensor(donor, acceptor)

    def to_tensors(self):
        """Converts traces into a list of tensors."""
        tensor_list = []
        for trace in self.traces:
            with tf.device('/CPU:0'):
                # Places this operation on CPU to reduce communication overhead.
                tensor_list.append(self.trace_to_tensor(trace))
        return tensor_list




"""Custom tensorflow layers."""

import tensorflow as tf
import tensorflow.keras as keras


@keras.utils.register_keras_serializable('Attention')
class Attention(keras.layers.Layer):
    def __init__(self, inner_dim=32, head_dim=2):
        super().__init__()
        self.inner_dim = inner_dim
        self.head_dim = head_dim
        
    def get_config(self):
        config = {
            "inner_dim": self.inner_dim,
            "head_dim": self.head_dim,
        }
        return config
    
    def build(self, input_shape):
        input_dim = int(input_shape[-1])
        self.q = self.add_weight("Q", shape=[self.head_dim, input_dim, self.inner_dim], initializer=tf.keras.initializers.GlorotUniform())
        self.k = self.add_weight("K", shape=[self.head_dim, input_dim, self.inner_dim], initializer=tf.keras.initializers.GlorotUniform())
        self.v = self.add_weight("V", shape=[self.head_dim, input_dim, self.inner_dim], initializer=tf.keras.initializers.GlorotUniform())
        self.linear = self.add_weight("linear", shape=[self.head_dim, self.inner_dim, input_dim])
    
    def get_weights(self, inputs, instruct_token):
        d = tf.cast(self.inner_dim, inputs.dtype)
        Q = tf.einsum('bti,hij->bhtj', inputs, self.q) 
        K = tf.einsum('bti,hij->bhtj', inputs, self.k)
        logits = tf.einsum('bhti,bhri->bhtr', Q, K) / tf.sqrt(d)  # same as paper, but do not understand
        if instruct_token is not None:
            one = tf.ones(shape=(1, ), dtype=Q.dtype)
            logits += tf.einsum('t,i,bhri->bhtr', one, instruct_token, K)
        weights = tf.nn.softmax(tf.einsum('bhti,bhri->bhtr', Q, K), axis=-1)
        return weights
    
    @tf.function
    def call(self, inputs, instruct_token=None):
        weights = self.get_weights(inputs, instruct_token)
        V = tf.einsum('bti,hij->bhtj', inputs, self.v)
        weighted_output = tf.einsum('bhri,bhtr->bhti', V, weights)
        output = inputs + tf.einsum('bhti,hij->btj', weighted_output, self.linear)
        normalized_output = tf.math.l2_normalize(output, axis=-1)
        return normalized_output


@keras.utils.register_keras_serializable('Attention')
class Conv(keras.layers.Layer):
    def __init__(self, width=8, inner_dim=32):
        super().__init__()
        self.inner_dim = inner_dim
        self.width = width
        self.stride = width // 2
        
    def get_config(self):
        config = {
            "inner_dim": self.inner_dim,
            "width": self.width,
        }
        return config
    
    def build(self, input_shape):
        input_dim = int(input_shape[-1])
        self.filter = self.add_weight("weight", shape=[self.width, input_dim, self.inner_dim], initializer=tf.keras.initializers.GlorotUniform())
        self.bias = self.add_weight("bias", shape=[1, 1, self.inner_dim], initializer=tf.zeros_initializer())
    
    @tf.function
    def call(self, inputs):
        output = tf.nn.conv1d(
            input=inputs,
            filters=self.filter,
            stride=self.stride,
            padding='SAME',
        )
        output -= self.bias
        return output
    
    def reconstruct(self, inputs):
        output = tf.repeat(inputs, self.stride, axis=-2)
        return output


@keras.utils.register_keras_serializable('Attention')
class Reconstructor(keras.layers.Layer):
    def __init__(self, width=8):
        super().__init__()
        self.width = width
        self.stride = width // 2

    def get_config(self):
        config = {
            "width": self.width,
        }
        return config
    
    def build(self, input_shape):
        input_dim = int(input_shape[-1])
        self.input_dim = input_dim
        self.remap = self.add_weight("remap", shape=[input_dim, self.stride, input_dim], initializer=tf.zeros_initializer())
    
    @tf.function
    def call(self, inputs):
        base = tf.repeat(inputs, self.stride, axis=-2)
        
        z = tf.einsum('iTh,htz->iTtz', inputs, self.remap)
        shape = tf.shape(z)
        new_shape = (shape[0], shape[1] * shape[2], shape[3])
        z = tf.reshape(z, new_shape)
                
        output = base + z
        output = tf.math.l2_normalize(output, axis=-1)
        return z


@keras.utils.register_keras_serializable('Attention')
class Summary(keras.layers.Layer):
    def __init__(self, inner_dim=16, outer_dim=16, head_dim=8):
        super().__init__()
        self.inner_dim = inner_dim
        self.outer_dim = outer_dim
        self.head_dim = head_dim
        
    def get_config(self):
        config = {
            "inner_dim": self.inner_dim,
            "outer_dim": self.outer_dim,
            "head_dim": self.head_dim,
        }
        return config
    
    def build(self, input_shape):
        input_dim = int(input_shape[-1])
        self.k = self.add_weight("K", shape=[self.head_dim, input_dim], initializer=tf.keras.initializers.GlorotUniform())
        self.v = self.add_weight("V", shape=[self.head_dim, input_dim, self.outer_dim], initializer=tf.keras.initializers.GlorotUniform())
        self.b = self.add_weight("B", shape=(1, self.outer_dim), initializer=tf.zeros_initializer())
    
    @tf.function
    def call(self, inputs):
        k = tf.math.l2_normalize(self.k, axis=-1)
        v = self.v
        inputs = tf.math.l2_normalize(inputs, axis=-1)
        w = tf.einsum("hi,bti->bht", k, inputs) / tf.sqrt(tf.cast(self.inner_dim, inputs.dtype))
        w = tf.nn.softmax(w, axis=-1)
        logits = tf.einsum("bht,hij,bti->bj", w, v, inputs) + self.b
        return logits


@keras.utils.register_keras_serializable('Attention')
class PrependTaskToken(tf.keras.layers.Layer):
    """Keras layer for prepending a learnable token to the sequence."""
    def __init__(self, token_dim, dtype=tf.float32):
        super().__init__()
        self.token_dim = token_dim
    
    def build(self, input_shape):
        self.token = self.add_weight("token", shape=[1, 1, self.token_dim], initializer=tf.keras.initializers.GlorotUniform())
        
    def get_config(self):
        config = {
            "token_dim": self.token_dim,
        }
        return config
    
    def call(self, seq):
        batch_size = tf.shape(seq)[0]
        token_tiled = tf.repeat(self.token, batch_size, axis=0)
        return tf.concat([token_tiled, seq], axis=1)


@keras.utils.register_keras_serializable('Attention')    
class Embedding(keras.layers.Layer):
    """Full e2e embedding layer for trace sequence analysis."""
    def __init__(self, conv, transformer, prepend_task_token, reconstructor=None, framewise=False):
        super().__init__()
        self.conv = conv
        self.transformer = transformer
        self.prepend_task_token = prepend_task_token
        self.framewise = framewise
        if reconstructor is not None:
            self.reconstructor = reconstructor
        else:
            self.reconstructor = conv.reconstruct
        
    def get_config(self):
        config = {
            "conv": self.conv,
            "transformer": self.transformer,
            "prepend_task_token": self.prepend_task_token,
            "reconstructor": self.reconstructor,
            "framewise": self.framewise,
        }
        return config
    
    def call(self, seq):
        local_features_input = self.conv(seq)
        # seq_input = tf.math.l2_normalize(self.prepend_task_token(local_features_input), axis=-1)
        seq_input = self.prepend_task_token(local_features_input)
        seq_output = self.transformer(seq_input)
        
        core_summary = seq_output[:, 0, :]
        core_seq = self.reconstructor(seq_output[:, 1:, :])
        
        if self.framewise:
            return core_seq
        else:
            return core_summary

    def get_visualization_token(self, seq):
        "Gets the first frame level token for projector visualization."
        local_features_input = self.conv(seq)
        # seq_input = tf.math.l2_normalize(self.prepend_task_token(local_features_input), axis=-1)
        seq_input = self.prepend_task_token(local_features_input)
        seq_output = self.transformer(seq_input)
        return seq_output[:, 1, :]


@keras.utils.register_keras_serializable('Attention')    
class Transformer(keras.layers.Layer):
    """Core transformer component for tokenized sequential data."""
    def __init__(self, depth, num_heads, inner_dim, use_layer_norm, dense_width_array, dense_activation, dropout):
        super().__init__()
        self.depth = depth
        self.num_heads = num_heads
        self.inner_dim = inner_dim
        self.use_layer_norm = use_layer_norm
        self.dense_width_array = dense_width_array
        self.dense_activation = dense_activation
        self.dropout = dropout

        self.position_embedding = PositionEmbedding()
        if self.use_layer_norm:
            self.layer_norm_layers = [keras.layers.LayerNormalization() for _ in range(2 * self.depth)]

        self.attention_layers = [keras.layers.MultiHeadAttention(num_heads=self.num_heads, key_dim=self.inner_dim) for _ in range(self.depth)]
        self.attention_dropout = [keras.layers.Dropout(self.dropout) for _ in range(self.depth)]

        self.dense_layers = []
        self.dense_dropout = []
        for _ in range(self.depth):
            self.dense_layers.append([])
            self.dense_dropout.append([])
            for dense_unit in [*self.dense_width_array, self.inner_dim]:
                self.dense_layers[-1].append(keras.layers.Dense(units=dense_unit, activation=self.dense_activation))
                self.dense_dropout[-1].append(keras.layers.Dropout(self.dropout))

    def get_config(self):
        config = {
            "depth": self.depth,
            "num_heads": self.num_heads,
            "inner_dim": self.inner_dim,
            "use_layer_norm": self.use_layer_norm,
            "dense_width_array": self.dense_width_array,
            "dense_activation": self.dense_activation,
            "dropout": self.dropout,
        }
        return config
        
    def call(self, base_seq_input):
        seq_input = base_seq_input + self.position_embedding(base_seq_input)
        for depth in range(self.depth):
            seq_input_residule = seq_input
            
            if self.use_layer_norm:
                ln = self.layer_norm_layers[2 * depth]
                seq_input = ln(seq_input)

            attention = self.attention_layers[depth]
            dropout = self.attention_dropout[depth]
            seq_input = dropout(attention(seq_input, seq_input))

            seq_input += seq_input_residule
            seq_input_residule = seq_input

            if self.use_layer_norm:
                ln = self.layer_norm_layers[2 * depth + 1]
                seq_input = ln(seq_input)

            for dense_ff, dropout in zip(self.dense_layers[depth], self.dense_dropout[depth]):
                seq_input = dense_ff(seq_input)
                seq_input = dropout(seq_input)
                
            seq_input += seq_input_residule

            if not self.use_layer_norm:
                seq_input = tf.math.l2_normalize(seq_input, axis=-1)
        return seq_input


@keras.utils.register_keras_serializable('Attention') 
class PositionEmbedding(tf.keras.layers.Layer):
    """Creates a positional embedding."""

    def __init__(self,
               max_length=400,
               initializer="glorot_uniform",
               seq_axis=1,
               **kwargs):

        super().__init__(**kwargs)
        if max_length is None:
            raise ValueError(
              "`max_length` must be an Integer, not `None`."
            )
        self._max_length = max_length
        self._initializer = tf.keras.initializers.get(initializer)
        self._seq_axis = seq_axis

    def get_config(self):
        config = {
            "max_length": self._max_length,
            "initializer": tf.keras.initializers.serialize(self._initializer),
            "seq_axis": self._seq_axis,
        }
        return config

    def build(self, input_shape):
        dimension_list = input_shape.as_list()
        width = dimension_list[-1]
        weight_sequence_length = self._max_length

        self._position_embeddings = self.add_weight(
            "embeddings",
            shape=[weight_sequence_length, width],
            initializer=self._initializer)

        super().build(input_shape)

    def call(self, inputs):
        input_shape = tf.shape(inputs)
        actual_seq_len = input_shape[self._seq_axis]
        position_embeddings = self._position_embeddings[:actual_seq_len, :]
        new_shape = [1 for _ in inputs.get_shape().as_list()]
        new_shape[self._seq_axis] = actual_seq_len
        new_shape[-1] = position_embeddings.get_shape().as_list()[-1]
        position_embeddings = tf.reshape(position_embeddings, new_shape)
        return tf.broadcast_to(position_embeddings, input_shape)


import scipy.io
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors
import tensorflow as tf
import smfret.trace_statistics

TRACE_KEY = 'traces'
DONOR_TO_ACCEPTOR_CORRECTION = 0.09
TIME_DENOMINATOR = 2000
FRET_EFFECTIVE_RANGE = (-0.2, 1.2)


class FRETTrace:
    """The base class for a smFRET trace."""
    def __init__(self, donor, acceptor, time, label):
        self.donor = donor
        self.acceptor = acceptor
        self.total = self.donor + self.acceptor
        self.time = time
        self.label = label
        self.statistics = None

    @property
    def fret(self):
        """Getter function for calculating the FRET value."""
        corrected_acceptor = self.acceptor - DONOR_TO_ACCEPTOR_CORRECTION * self.donor
        fret = (corrected_acceptor) / (self.donor + corrected_acceptor + np.finfo(np.float32).eps)
        return fret
    
    def analyze(self):
        self.statistics = smfret.trace_statistics.analyze_trace(self)

        
class FRETTraceSet:
    """The base class for a dataset to store traces."""
    def __init__(self):
        self.original_file = None
        self.time = None
        self.donor = None
        self.acceptor = None
        self.size = None
        self.label = None
        self.is_labeled = None
        self._cached_stacked_time = None
        self.traces = []

    def plot_traces(self, n_start=0, n_traces=10, selected_traces_only=False):
        """Plots samples of stored traces in the dataset."""
        i = n_start - 1
        count = 0
        while count < n_traces and i < (self.size - 1):
            i += 1
            if selected_traces_only and not any(self.label[i]):
                continue
            count += 1
            plt.subplot(n_traces, 1, count)
            plt.plot(self.time, self.donor[i, :], color='tab:blue', linewidth=0.5)
            plt.plot(self.time, self.acceptor[i, :], color='tab:red', linewidth=0.5)
            
            start_end_frames = []
            max_line = max(np.max(self.donor[i, :]), np.max(self.acceptor[i, :]))
            if len(self.label[i, :]) == len(self.donor[i, :]):
                for j in range(self.time.shape[0]):
                    last_frame_label = self.label[i, j - 1] if j > 0 else 0
                    if np.logical_xor(last_frame_label, self.label[i, j]):
                        plt.axvline(x=self.time[j], color='tab:green', linestyle='-.')
                        start_end_frames.append(j)
                        if len(start_end_frames) == 2:
                            plt.plot(
                                start_end_frames,
                                [max_line, max_line],
                                color='tab:green',
                                linewidth=2,
                            )
                            start_end_frames = []
                if len(start_end_frames) == 1:
                    # Selection continues to the last frame.
                    plt.axvline(x=self.time[-1], color='tab:green', linestyle='-.')
                    start_end_frames.append(len(self.time) - 1)
                    plt.plot(
                        start_end_frames,
                        [max_line, max_line],
                        color='tab:green',
                        linewidth=2,
                    )
                    
            self.traces[i].analyze()
            if self.traces[i].statistics.trace_lifetime is not None:
                plt.axvline(x=self.traces[i].statistics.trace_lifetime, color='tab:orange', linestyle=':')
            if self.traces[i].statistics.mean_donor is not None:
                plt.axhline(y=self.traces[i].statistics.mean_donor, color='tab:blue', linestyle=':')
            if self.traces[i].statistics.mean_acceptor is not None:
                plt.axhline(y=self.traces[i].statistics.mean_acceptor, color='tab:red', linestyle=':')
            
            plt.annotate(f'trace {i}', (len(self.time) / 2.0, 0.9))
            plt.ylabel('Intensity (A.U.)')
        plt.xlabel('Frame')
        plt.tight_layout()

    def __str__(self):
        string = f"""FRET traces dataset with {self.size} traces.
size = {self.size}
length = {self.time.shape[0]}
is_labeled = {self.is_labeled}
        """
        return string

    def __repr__(self):
        return self.__str__()
    
    def __add__(self, other):
        return CombinedTraceSet(self, other)
    
    def fret_prob_distribution(self, bins=28, predicted_label=None):
        """Calculates the prob distribution of FRET values."""
        if predicted_label is None:
            idx = np.where(self.label)
        else:
            idx = np.where(predicted_label)
        fret = self.fret[idx]
        prob, bins = np.histogram(fret, bins=bins, range=FRET_EFFECTIVE_RANGE, density=True)
        return prob * (bins[1] - bins[0]), bins
    
    def valid_fret_values(self, predicted_label=None):
        """Calculates the FRET values of selected segments."""
        if predicted_label is None:
            idx = np.where(self.label)
        else:
            idx = np.where(predicted_label)
        valid_fret = self.fret[idx]
        cleaned_fret = valid_fret[(valid_fret > FRET_EFFECTIVE_RANGE[0]) 
                                  & (valid_fret < FRET_EFFECTIVE_RANGE[1])]
        return cleaned_fret

    def plot_fret_histogram_of_selected_segments(self, bins=28, predicted_label=None):
        """Plots the histogram of FRET value for selected segments."""
        idx = np.where(self.label)
        fret = self.fret[idx]
        plt.hist(fret, bins=bins, range=FRET_EFFECTIVE_RANGE, alpha=0.8, facecolor='tab:blue', edgecolor='k', linewidth=1.5, label='Ground Truth', density=True)
        if predicted_label is not None:
            idx = np.where(predicted_label)
            fret = self.fret[idx]
            plt.hist(fret, bins=bins, range=FRET_EFFECTIVE_RANGE, alpha=0.8, facecolor='tab:orange', edgecolor='k', linewidth=1.5, label='Prediction', density=True)
        plt.legend()
    
    def to_tensor(self, size=None, normalize=True):
        """Converts the smFRET data into the form for training TensorFlow models."""
        donor = tf.cast(self.donor, tf.float32)
        acceptor = tf.cast(self.acceptor, tf.float32)
        total = donor + acceptor
        tensor = tf.stack([donor, acceptor, total], axis=-1)
        total_max = tf.reduce_max(total, axis=(-1))
        
        if normalize:
            total_max = tf.expand_dims(tf.expand_dims(total_max, axis=-1), axis=-1)
            tensor = tensor / total_max
        
        if self._cached_stacked_time is None:
            time = tf.cast(tf.repeat([self.time], self.size, axis=0), dtype=tf.float32)
            time = tf.expand_dims(time / tf.cast(TIME_DENOMINATOR, time.dtype), axis=-1)
            self._cached_stacked_time = time
        else:
            time = self._cached_stacked_time
        
        tensor = tf.concat([tensor, time], axis=-1)
        return tensor
    
    @property
    def fret(self):
        """Getter function for calculating the FRET value."""
        corrected_acceptor = self.acceptor - DONOR_TO_ACCEPTOR_CORRECTION * self.donor
        fret = (corrected_acceptor) / (self.donor + corrected_acceptor + np.finfo(np.float32).eps)
        return fret
    
    @property
    def entire_trace_label(self):
        """Getter function for converting frame-wise labels into trace-wise labels."""
        trace_label = np.sum(self.label, axis=-1) > 0
        return trace_label
    
    def plot_fret_jump_heatmap(self, n_frames=1):
        """Plots a heatmap of the FRET values for current and n frames before."""
        plt.hist2d(
            self.fret[:, :-n_frames].flatten(),
            self.fret[:, n_frames:].flatten(),
            bins=50,
            range=((0, 1), (0, 1)),
            norm=matplotlib.colors.LogNorm(),
            cmap='plasma',
        )
        plt.colorbar()
        plt.xlabel('FRET after')
        plt.ylabel('FRET before')
    
    def broadcast_data_to_traces(self):
        """Broadcasts the data store in arrays into individual trace objects."""
        self.traces = []
        for i in range(self.size):
            trace = FRETTrace(
                time=self.time,
                donor=self.donor[i, :],
                acceptor=self.acceptor[i, :],
                label=self.label[i, :],
            )
            self.traces.append(trace)
        
    def analyze(self):
        """Runs the analyze() method for each smFRET trace."""
        for trace in self.traces:
            trace.analyze()
 
    def statistics_to_tensor(self):
        """Converts the statistics of each trace to a dense Tensor."""
        return np.array([trace.statistics for trace in self.traces])
    
    def copy(self, trace_set):
        """Copied data from another trace set."""
        self.original_file = trace_set.original_file
        self.time = trace_set.time
        self.donor = trace_set.donor
        self.acceptor = trace_set.acceptor
        self.size = trace_set.size
        self.label = trace_set.label
        self.is_labeled = trace_set.is_labeled
        self.traces = trace_set.traces
        
    def trim(self, n_frame, start_frame=0):
        """Shortens each trace to a fixed number of frames."""
        self.donor = self.donor[:, start_frame : start_frame + n_frame]
        self.acceptor = self.acceptor[:, start_frame : start_frame + n_frame]
        self.label = self.label[:, start_frame : start_frame + n_frame]
        self.time = self.time[:self.donor.shape[1]]
        self.broadcast_data_to_traces()
        
    def train_test_split(self, train_percentage):
        """Splits the trace set into two randomly selected sets."""
        idx = np.random.choice([True, False], size=self.size, p=[train_percentage, 1.0 - train_percentage])
        
        train_set = FRETTraceSet()
        train_set.original_file = self.original_file
        train_set.time = self.time
        train_set.donor = self.donor[idx, :]
        train_set.acceptor = self.acceptor[idx, :]
        train_set.size = np.sum(idx)
        train_set.label = self.label[idx, :]
        train_set.is_labeled = self.is_labeled
        train_set.broadcast_data_to_traces()
        
        test_set = FRETTraceSet()
        test_set.original_file = self.original_file
        test_set.time = self.time
        test_set.donor = self.donor[~idx, :]
        test_set.acceptor = self.acceptor[~idx, :]
        test_set.size = self.size - np.sum(idx)
        test_set.label = self.label[~idx, :]
        test_set.is_labeled = self.is_labeled
        test_set.broadcast_data_to_traces()
        
        return (train_set, test_set)

    def get_subset(self, indices):
        """Gets a subset of the trace set."""
        new_set = FRETTraceSet()
        new_set.original_file = self.original_file
        new_set.time = self.time
        new_set.donor = self.donor[indices, :]
        new_set.acceptor = self.acceptor[indices, :]
        new_set.size = new_set.donor.shape[0]
        new_set.label = self.label[indices, :]
        new_set.is_labeled = self.is_labeled
        new_set.broadcast_data_to_traces()
        return new_set
        
            

class MatlabTraceSet(FRETTraceSet):
    def __init__(self, file: str):
        super().__init__()
        matlab_data = scipy.io.loadmat(file)[TRACE_KEY]
        self.file = file
        self.original_file = matlab_data['file'][0, 0][0]
        self.time = matlab_data['time'][0, 0][0].astype(np.int64)
        self.donor = matlab_data['donor'][0, 0].astype(np.int64)
        self.acceptor = matlab_data['acceptor'][0, 0].astype(np.int64)
        self.size = matlab_data['count'][0, 0][0, 0].astype('int')
        self.label = matlab_data['label'][0, 0].astype(np.int64)
        self.is_labeled = matlab_data['islabeled'][0, 0][0, 0].astype('bool')
        self.broadcast_data_to_traces()
        

class CombinedTraceSet(FRETTraceSet):
    def __init__(self, set_a, set_b):
        super().__init__()
        if not np.array_equal(set_a.time, set_b.time):
            #TODO(leyou): Use padding when combining two datasets of different time length.
            raise ValueError("The time axis of two trace sets much be the same.")
        
        self.time = set_a.time
        self.donor = np.concatenate((set_a.donor, set_b.donor), axis=0)
        self.acceptor = np.concatenate((set_a.acceptor, set_b.acceptor), axis=0)
        self.size = set_a.size + set_b.size
        self.label = np.concatenate((set_a.label, set_b.label), axis=0)
        self.is_labeled = set_a.is_labeled or set_b.is_labeled
        self.broadcast_data_to_traces()
        
        


"""Library for finetuning pretrained models."""
import tensorflow as tf
import tensorflow.keras as keras
import numpy as np
from sklearn.linear_model import LogisticRegression


def encode_traces(trace_set, encoder):
    """Encodes traces into a numpy matrix."""
    with tf.device('CPU:0'):
        embedding = encoder.predict(trace_set.to_tensor(), verbose=0)
    return embedding
  

def finetune_classification_head(encoder, train_sets, epochs, num_classes=2, trace_level=True, lr=1e-3, verbose=0, return_embedding=False, use_sklearn=False, embedding_list=None, l2_regularization=0.0):
    """Finetunes a classification head."""
    if trace_level:
        if embedding_list is None:
            embedding_list = [encode_traces(trace_set, encoder) for trace_set in train_sets]
        label_list = [tf.cast(np.max(trace_set.label, axis=-1), tf.int64) for trace_set in train_sets]
    else:
        if embedding_list is None:
            embedding_list = [tf.reshape(encode_traces(trace_set, encoder), (-1, encoder.output_shape[-1])) for trace_set in train_sets]
        label_list = [tf.cast(tf.reshape(trace_set.label, (-1, )), tf.int64) for trace_set in train_sets]

    embedding = tf.concat(embedding_list, axis=0)
    label = np.concatenate(label_list, axis=0)

    if use_sklearn:
        clf = LogisticRegression(max_iter=5000, tol=1e-6).fit(embedding, label)
        if clf.coef_.shape[0] == 1:
            clf.coef_ = np.concatenate([np.zeros_like(clf.coef_), clf.coef_], axis=0)
            clf.intercept_ = np.concatenate([np.zeros_like(clf.intercept_), clf.intercept_], axis=0)
        num_classes = clf.coef_.shape[0]

    readout = keras.layers.Dense(num_classes, kernel_regularizer=keras.regularizers.l2(l2_regularization))

    readout_model = keras.Sequential([readout])
    readout_model.compile(
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=tf.keras.optimizers.legacy.Adam(lr),
    )

    #TODO: Customize class weights if necessary.
    class_weights = None
    if use_sklearn:
        readout.build(input_shape=(None, embedding.shape[-1]))
        readout.variables[0].assign(clf.coef_.T)
        readout.variables[1].assign(clf.intercept_)
    else:
        if trace_level:
            readout_model.fit(embedding, label, batch_size=256, epochs=100, verbose=verbose, class_weight=class_weights,)
        else:
            readout_model.fit(embedding, label, batch_size=1024, epochs=2, verbose=verbose, class_weight=class_weights,)

    input = encoder.input
    output = readout(encoder(input))
    classifier = keras.Model(inputs=input, outputs=output)
    
    if return_embedding:
        return classifier, embedding_list
    else:
        return classifier


"""Simulated smFRET traces dataset classes for multi-task learning."""

import numpy as np
from tqdm.auto import tqdm
import multiprocessing

import smfret.dataset
from smfret.dataset import MatlabTraceSet
from smfret.dataset import FRETTraceSet
from smfret.dataset import FRETTrace
from smfret.trace_simulator import Simulator
from smfret.trace_simulator import SingleChannelSimulator
from smfret.trace_simulator import ParameterGenerator
from smfret.trace_simulator import SimulatedTraceSet

import tensorflow as tf
from datetime import datetime as tm

rng = np.random.default_rng()

class EvolvingTraceSet(SimulatedTraceSet):
    """Class for trace sets capable of replacing X% traces with new traces."""
    def __init__(self, size, params_gen):
        super().__init__(None)
        self.size = 0
        self.params_gen = params_gen
        self.traces = []
        self.populate(size)
        self.vectorize_traces()
        self.is_labeled = True
    
    def populate(self, n: int):
        """Populates the dataset using the simulator."""
        new_traces = []
        try:
            pos = multiprocessing.current_process()._identity[0] - 1
            disable = False
        except:
            disable = True
            pos = 0
            
        desc_text = self.__class__.__name__
        for _ in tqdm(range(n), desc=desc_text, position=pos, disable=disable):
            new_traces.append(self.trace_gen())
        self.traces += new_traces
        self.size += n

    def delete(self, n: int):
        """Removes the traces on the top of the traces list."""
        self.traces = self.traces[n:]
        self.size -= n
        
    def evolve(self, n):
        """Adds new traces and removes old traces."""
        self.populate(n)
        self.delete(n)
        self.vectorize_traces()

    def vectorize_traces(self):
        """Stacks traces signals into np vectors."""
        self.time = self.traces[0].time
        self.donor = np.stack([trace.donor for trace in self.traces], axis=0)
        self.acceptor = np.stack([trace.acceptor for trace in self.traces], axis=0)
        self.label = np.stack([trace.label for trace in self.traces], axis=0)
    
    def trace_gen(self):
        """Generates a single labeled trace."""
        pass
    
    def save(self, file):
        """Saves the dataset to file."""
        self.vectorize_traces()
        data = {
            'time': self.time,
            'donor': self.donor,
            'acceptor': self.acceptor,
            'label': self.label,
            'size': self.size
        }
        np.savez(file, **data)
    

class SavedTraceSet(EvolvingTraceSet):
    """Class for trace sets saved on disk."""
    def __init__(self, size, file):
        self.file = file
        self.counter = 0
        self.epochs = 0
        self.cached_traces_data = None
        self.load_file_to_cache()
        super().__init__(size=size, params_gen=None)
        
    def resize(self, size):
        """Changes the size of the dataset."""
        self.counter = 0
        self.epochs = 0
        super().__init__(size, params_gen=None)
        
    def load_file_to_cache(self):
        """Reads the saved traces."""
        self.cached_traces_data = dict(np.load(self.file))
        total = self.cached_traces_data['donor'] + self.cached_traces_data['acceptor']
        max_total = np.expand_dims(total.max(axis=-1), axis=-1)
        self.cached_stacked = tf.cast(np.stack([
            self.cached_traces_data['donor'] / max_total,
            self.cached_traces_data['acceptor'] / max_total,
            total / max_total,
            np.repeat([self.cached_traces_data['time']], self.cached_traces_data['size'], axis=0) / smfret.dataset.TIME_DENOMINATOR
        ], axis=-1), tf.bfloat16)

        # reduce the memory footprint of cached_traces_data
        for key in self.cached_traces_data:
            if self.cached_traces_data[key].dtype == np.float64:
                self.cached_traces_data[key] = self.cached_traces_data[key].astype(np.float32, casting='same_kind')
        
    
    def trace_gen(self):
        trace = FRETTrace(
            donor=self.cached_traces_data['donor'][self.counter, :],
            acceptor=self.cached_traces_data['acceptor'][self.counter, :],
            time=self.cached_traces_data['time'],
            label=self.cached_traces_data['label'][self.counter, ...],
        )
        trace.counter = self.counter
        
        self.counter += 1
        if self.counter == self.cached_traces_data['size']:
            self.counter = 0
            self.epochs += 1
            
        return trace
    
    def vectorize_traces(self):
        """Overwrites the default behavior of vectorize_traces to gain performance."""
        with tf.device('/CPU:0'):
            self.time = self.traces[0].time
            counters = [trace.counter for trace in self.traces]
            self.donor = self.cached_traces_data['donor'][counters, :]
            self.acceptor = self.cached_traces_data['acceptor'][counters, :]
            self.label = self.cached_traces_data['label'][counters, ...]
            self.stacked = tf.gather(self.cached_stacked, counters)
        
    
    def to_tensor(self, size=None, normalize=True):
        """Converts the smFRET data into the form for training TensorFlow models."""
        with tf.device('/CPU:0'):
            return tf.cast(self.stacked, tf.float32)


class FRETStateTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET states."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        trace.label = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        return trace


class TwoStateQuickDynamicTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with quick FRET transitions."""
    fret_states_gap_min = 0.5
    lifetime_min = 500
    transition_prob_min = 0.03
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            if params.num_states != 2:
                continue
            elif (np.max(params.fret_states) - np.min(params.fret_states)) < self.fret_states_gap_min:
                continue
            elif min(params.transition_prob_matrix[0, 1], params.transition_prob_matrix[1, 0]) < self.transition_prob_min:
                continue
            trace = Simulator(params).generate()
        return trace
    
class TwoStateQuickDynamicNBDTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with the N_(b+d)."""
    fret_states_resolution = 0.1
    fret_states_gap_min = 0.5
    lifetime_min = 500
    transition_prob_min = 0.05
    lifetime_min = 500
    quanta = 50
    lower_limit = 1
    upper_limit = 400
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            if params.num_states != 2:
                continue
            elif (np.max(params.fret_states) - np.min(params.fret_states)) < self.fret_states_gap_min:
                continue
            elif min(params.transition_prob_matrix[0, 1], params.transition_prob_matrix[1, 0]) < self.transition_prob_min:
                continue
            trace = Simulator(params).generate()

        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        nbd = np.sum(states[1:] != states[:-1])
        trace.label = np.digitize(nbd, self.quantize_bins, right=True)
        return trace


class TwoStateQuickDynamicTransitionProbTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with the transition probability."""
    fret_states_resolution = 0.1
    fret_states_gap_min = 0.5
    lifetime_min = 500
    transition_prob_min = 0.05
    lifetime_min = 500
    quanta = 4
    lower_limit = 1
    upper_limit = 20
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            if params.num_states != 2:
                continue
            elif (np.max(params.fret_states) - np.min(params.fret_states)) < self.fret_states_gap_min:
                continue
            elif min(params.transition_prob_matrix[0, 1], params.transition_prob_matrix[1, 0]) < self.transition_prob_min:
                continue
            trace = Simulator(params).generate()
            
        if params.fret_states[0] > params.fret_states[1]:
            inverse_p = 1. /  params.transition_prob_matrix[0, 1]
        else:
            inverse_p = 1. / params.transition_prob_matrix[1, 0]

        trace.label = np.digitize(inverse_p, self.quantize_bins, right=True)
        return trace
    

class SelectedSegmentsLowFRETTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with selected segments."""
    lifetime_min = 500
    fret_state_max = 0.2
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            if params.num_states != 1:
                continue
            elif np.min(params.fret_states) > self.fret_state_max:
                continue
            trace = Simulator(params).generate()
        return trace
    
    
class HighFRETBinaryTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET states."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        trace.label = np.array((states == np.max(states[:trace.acceptor_lifetime])))
        trace.label[trace.acceptor_lifetime:] = 0
        return trace
    
    
class LowFRETBinaryTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET states."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        trace.label = np.array((states == np.min(states[:trace.acceptor_lifetime])))
        trace.label[trace.acceptor_lifetime:] = 0
        return trace
    
    
class FRETStateHighestTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET states."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        trace.label = np.max(states[:trace.acceptor_lifetime])
        return trace


class FRETStateLowestTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET states."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        trace.label = np.min(states[:trace.acceptor_lifetime])
        return trace
    

class FRETStateCountTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with the number of FRET states."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    quanta = 1
    lower_limit = 1
    upper_limit = 4
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        trace.label = np.digitize(np.unique(states).size, self.quantize_bins, right=True)
        return trace
    
    
class NBDTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with the N_(b+d)."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    quanta = 50
    lower_limit = 1
    upper_limit = 400
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        nbd = np.sum(states[1:] != states[:-1])
        trace.label = np.digitize(nbd, self.quantize_bins, right=True)
        return trace
    
    
class DwellTimeTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with the average dwell time."""
    fret_states_resolution = 0.1
    lifetime_min = 500
    quanta = 20
    lower_limit = 10
    upper_limit = 200
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()
        
        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        current_dwell_time = 0
        last_state = None
        dwell_times = []
        for state in states:
            if state == last_state:
                current_dwell_time += 1
            else:
                dwell_times.append(current_dwell_time)
                current_dwell_time = 1
                last_state = state
            if state == 0:
                break
        
        trace.label = np.digitize(np.mean(dwell_times), self.quantize_bins, right=True)
        return trace
    
    
class TransitionRateTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with the rate of state transitions."""
    fret_states_resolution = 0.2
    lifetime_min = 500
    quanta = np.log(2.0)
    lower_limit = np.log(10.0)
    upper_limit = np.log(200.0)
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    eps = 1e-8
    
    def trace_gen(self):
        while True:
            params = self.params_gen.generate()
            fret_states = np.digitize(params.fret_states, np.arange(0, 1, self.fret_states_resolution), right=True)
            transition_out_rate = np.zeros_like(params.fret_states)
            
            if len(np.unique(fret_states)) <= 1:
                continue

            for i, fret in enumerate(params.fret_states):
                state = fret_states[i]
                for j, prob in enumerate(params.transition_prob_matrix[i, :]):
                    other_state = fret_states[j]
                    if other_state != state:
                        transition_out_rate[i] += prob
            
            avg_transition_time = np.mean(1.0 / (transition_out_rate + self.eps))
            log_transition_time = np.log(avg_transition_time)
            label = [log_transition_time]
            
            trace = Simulator(params).generate()

            if trace.acceptor_lifetime <= self.lifetime_min:
                continue
            else:
                trace.label = np.digitize(label, self.quantize_bins, right=True)
                return trace
            
            
class NoiseLevelTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with noise levels."""
    lifetime_min = 500
    quanta = 2
    lower_limit = 1
    upper_limit = 10
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()

        snr_signal = min(params.snr_signal, params.snr_background)
        trace.label = np.digitize(snr_signal, self.quantize_bins, right=True)
        return trace
    

class DonorLifetimeTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with donor lifetime."""
    quanta = 400.0
    lower_limit = 0.0
    upper_limit = 2000.0
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        params = self.params_gen.generate()
        trace = Simulator(params).generate()
        lifetime = trace.donor_lifetime
        trace.label = np.digitize(lifetime, self.quantize_bins, right=True)
        return trace
    
    
class AcceptorLifetimeTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with acceptor lifetime."""
    quanta = 400.0
    lower_limit = 0.0
    upper_limit = 2000.0
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        params = self.params_gen.generate()
        trace = Simulator(params).generate()
        lifetime = trace.acceptor_lifetime
        trace.label = np.digitize(lifetime, self.quantize_bins, right=True)
        return trace
    
    
class MultistepPhotobleachingTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with multistep photobleaching."""
    lifetime_min = 200
    quanta = 1.0
    lower_limit = 1.0
    upper_limit = 4.0
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        params = self.params_gen.generate()
        traces = []
        n_steps = rng.integers(self.lower_limit, self.upper_limit, endpoint=True)
        while len(traces) < n_steps:
            trace = Simulator(params).generate()
            lifetime = trace.acceptor_lifetime
            if lifetime is None or lifetime < self.lifetime_min:
                continue
            traces.append(trace)
        trace = traces[0]
        trace.donor = np.mean([x.donor for x in traces], axis=0)
        trace.acceptor = np.mean([x.acceptor for x in traces], axis=0)
        label = [n_steps]
        trace.label = np.digitize(label, self.quantize_bins, right=True)
        return trace
    
    
class SelectedSegmentsTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with selected segments."""
    lifetime_min = 500
    
    def trace_gen(self):
        params = self.params_gen.generate()
        while True:
            trace = Simulator(params).generate()
            lifetime = trace.donor_lifetime
            if lifetime is None or lifetime < self.lifetime_min:
                continue
            else:
                break
        return trace
    
    
class ReemergentTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with reemmergent trace segments."""
    lifetime_min = 200
    lifetime_max = 600
    
    quanta = 1.0
    lower_limit = 1.0
    upper_limit = 4.0
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    gap_min = 200
    gap_max = 500
    
    def __init__(self, size, params_gen, binary_label=False):
        self.binary_label = binary_label
        super().__init__(size, params_gen)
    
    def trace_gen(self):
        params = self.params_gen.generate()
        traces = []
        n_traces = rng.integers(self.lower_limit, self.upper_limit, endpoint=True)
        
        while True:
            trace = Simulator(params).generate()
            lifetime = trace.acceptor_lifetime
            if lifetime is None or lifetime < self.lifetime_min or lifetime > self.lifetime_max:
                continue
            traces.append(trace)
            if len(traces) == n_traces:
                break
                
        combined_trace = None
        trace_length = traces[0].time.size
        
        for (i, trace) in enumerate(traces):
            # Concats all traces
            lifetime = trace.acceptor_lifetime
            if i < n_traces - 1:
                gap = rng.integers(self.gap_min, self.gap_max, endpoint=True)
                max_frame = lifetime + gap
            else:
                max_frame = None
            trace.donor = trace.donor[:max_frame]
            trace.acceptor = trace.acceptor[:max_frame]
            if self.binary_label:
                if i == 0:
                    trace.label = trace.label[:max_frame]
                else:
                    trace.label = 0 * trace.label[:max_frame]
            else:
                trace.label = (i + 1) * trace.label[:max_frame]
            
            if combined_trace is None:
                combined_trace = trace
            else:
                combined_trace.donor = np.concatenate([combined_trace.donor, trace.donor])
                combined_trace.acceptor = np.concatenate([combined_trace.acceptor, trace.acceptor])
                combined_trace.label = np.concatenate([combined_trace.label, trace.label])
        
        combined_trace.donor = combined_trace.donor[:trace_length]
        combined_trace.acceptor = combined_trace.acceptor[:trace_length]
        combined_trace.total = combined_trace.donor + combined_trace.acceptor
        combined_trace.label = combined_trace.label[:trace_length]
        combined_trace.time = np.arange(trace_length)    
        
        return combined_trace
    
    
class BackgroundTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET background intensity."""
    resolution = 0.02
    background_clip = 0.2
    lifetime_min = 500
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = Simulator(params).generate()
        
        background = params.background / params.total_intensity
        states = np.digitize(background, np.arange(0, self.background_clip, self.resolution), right=True)
        trace.label = states
        return trace
    
    
class CoincidentTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with coincidentally overlapping donor and acceptor."""
    lifetime_min = 500
    
    def trace_gen(self):
        traces = []
        while len(traces) < 2:
            trace = None
            while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
                params = self.params_gen.generate()
                trace = Simulator(params).generate()
            traces.append(trace)
        coincident_happened = rng.uniform(0, 1) > 0.5
        if coincident_happened:
            traces[0].donor = traces[1].donor  # replaces donor
        trace = traces[0]
        trace.label = coincident_happened
        return trace


class SingleChannelStateTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET states."""
    fret_states_resolution = 0.1
    lifetime_min = 200
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = SingleChannelSimulator(params).generate()

        normalized_state = trace.acceptor_ideal / np.max(trace.acceptor_ideal)
        trace.label = np.digitize(normalized_state, np.arange(0, 1, self.fret_states_resolution), right=True)
        return trace


class SingleChannelStateFastTransitionTraceSet(SingleChannelStateTraceSet):
    """Class for trace sets labeled with FRET states for traces with fast transitions."""
    fret_states_resolution = 0.1
    transition_prob_increase_factor = 5
    snr_increase_factor = 2
    lifetime_min = 200

    def __init__(self, *args, **kwargs):
        """Overrides the parameter generator."""
        super().__init__(*args, **kwargs)
        self.params_gen = self.params_gen.copy()  # makes it safe to mutate the generator
        self.params_gen.transition_prob_fn = lambda: self.transition_prob_increase_factor * self.params_gen.transition_prob_fn()
        self.params_gen.snr_signal_fn = lambda: self.snr_increase_factor * self.params_gen.snr_signal_fn()
        self.params_gen.snr_background_fn = lambda: self.snr_increase_factor * self.params_gen.snr_background_fn()


class SingleChannelDwellTimeTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with the average dwell time."""
    fret_states_resolution = 0.2
    lifetime_min = 500
    quanta = 20
    lower_limit = 10
    upper_limit = 400
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            trace = SingleChannelSimulator(params).generate()
        
        states = np.digitize(trace.fret_ideal, np.arange(0, 1, self.fret_states_resolution), right=True)
        
        current_dwell_time = 0
        last_state = None
        dwell_times = []
        for state in states:
            if state == last_state:
                current_dwell_time += 1
            else:
                dwell_times.append(current_dwell_time)
                current_dwell_time = 1
                last_state = state
            if state == 0:
                break
        
        trace.label = np.digitize(np.mean(dwell_times), self.quantize_bins, right=True)
        return trace


class SingleChannelBlinkingTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with FRET states."""
    lifetime_min = 200
    blinking_rate_increase_factor = 10
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            params.non_cy3_blink_lifetime /= self.blinking_rate_increase_factor
            trace = SingleChannelSimulator(params).generate()

        trace.label = (trace.acceptor_ideal == 0) & (trace.time <= trace.acceptor_lifetime)
        return trace


class SingleChannelLifetimeTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with lifetime."""
    quanta = 200.0
    lower_limit = 0.0
    upper_limit = 2000.0
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)
    
    def trace_gen(self):
        params = self.params_gen.generate()
        trace = SingleChannelSimulator(params).generate()
        lifetime = trace.acceptor_lifetime
        trace.label = np.digitize(lifetime, self.quantize_bins, right=True)
        return trace


class SingleChannelNoiseLevelTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with noise levels."""
    lifetime_min = 500
    quanta = 2
    lower_limit = 4
    upper_limit = 20
    quantize_bins = np.arange(lower_limit, upper_limit, quanta)

    def snr_increase_factor_fn(self):
        """Generates a random factor to increase the SNR."""
        return rng.uniform(1.0, 4.0)
    
    def trace_gen(self):
        trace = None
        while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
            params = self.params_gen.generate()
            params.snr_signal *= self.snr_increase_factor_fn()
            params.snr_background *= self.snr_increase_factor_fn()
            trace = SingleChannelSimulator(params).generate()

        effective_snr = min(params.snr_signal, params.snr_background) / params.total_intensity * np.max(trace.acceptor_ideal)  # removes the contribution from donor
        trace.label = np.digitize(effective_snr, self.quantize_bins, right=True)
        return trace


class SingleChannelPhotobleachingStepsFrameLevelTraceSet(EvolvingTraceSet):
    """Class for trace sets labeled with photobleaching steps at the frame level."""
    num_steps_max = 6
    num_steps_min = 1
    lifetime_min = 200

    def num_steps_fn(self):
        """Generates the number of photobleaching steps."""
        return rng.integers(low=self.num_steps_min, high=self.num_steps_max, endpoint=True)


    def params_override(self, params):
        """Overrides params for this particular trace set."""
        params.num_states = 1
        params.initial_state = 0
        params.transition_prob_matrix = np.array([[1.0]])
        params.fret_states=[1.0]
        return params
    
    def trace_gen(self):

        traces = []
        num_steps = self.num_steps_fn()
        for _ in range(num_steps):
            trace = None
            while trace is None or trace.acceptor_lifetime <= self.lifetime_min:
                params = self.params_override(self.params_gen.generate())
                trace = SingleChannelSimulator(params).generate()

            trace.label = trace.time <= trace.acceptor_lifetime
            traces.append(trace)

        trace = traces[0]
        trace.acceptor = np.mean([x.acceptor for x in traces], axis=0)
        trace.acceptor_ideal = np.mean([x.acceptor_ideal for x in traces], axis=0)
        trace.label = np.sum([x.label for x in traces], axis=0)

        trace.label -= trace.label[-1]  # removes the unfinished steps from counting

        return trace


class SingleChannelPhotobleachingStepsTraceLevelTraceSet(SingleChannelPhotobleachingStepsFrameLevelTraceSet):
    """Class for trace sets labeled with photobleaching steps at the trace level."""
    def trace_gen(self):
        """Overrides the frame level trace gen."""
        trace = super().trace_gen()
        trace.label = trace.label[0] - trace.label[-1]
        return trace


"""Library for rendering META-SiM Atlas."""
import os
import requests
import zipfile
import shutil
import joblib


FILENAME = 'atlas_pp.joblib'
FILEPATH = os.path.join(os.path.dirname(__file__), FILENAME)  # Path in the same directory as atlas.py
FILE_URL = "https://storage.googleapis.com/fret_traces/Atlas/atlas_pp.joblib.zip"
DENSITY_MODEL_FILENAME = 'atlas_pp_density_models.joblib'

DENSITY_MODEL_FILEPATH = os.path.join(os.path.dirname(__file__), DENSITY_MODEL_FILENAME)  # Path in the same directory as atlas.py

# Constant for Atlas category names.
SORTED_NAMES = [
 '1-c-l-s',
 '1-n-l-s',
 '1-c-m-s',
 '1-n-m-s',
 '1-c-h-s',
 '1-n-h-s',
 '2-c-lm-s',
 '2-n-lm-s',
 '2-c-mh-s',
 '2-n-mh-s',
 '2-c-lh-s',
 '2-n-lh-s',
 '2-c-lm-f',
 '2-n-lm-f',
 '2-c-mh-f',
 '2-n-mh-f',
 '2-c-lh-f',
 '2-n-lh-f',
 '3-c-lmh-s',
 '3-n-lmh-s',
 '3-c-lmh-f',
 '3-n-lmh-f',
]
# Shorter names for 1-state categories.
CLEANED_NAMES = {
 '1-c-l-s': '1-c-l',
 '1-n-l-s': '1-n-l',
 '1-c-m-s': '1-c-m',
 '1-n-m-s': '1-n-m',
 '1-c-h-s': '1-c-h',
 '1-n-h-s': '1-n-h',
}


def download_and_unzip_atlas(file_url=FILE_URL, filename=FILENAME):
    """
    Downloads and unzips the Atlas file if it doesn't exist.

    Args:
        file_url: The URL of the atlas.pp.zip file.
        filename: The desired filename for the unzipped file.
    """

    filepath = os.path.join(os.path.dirname(__file__), filename)  # Path in the same directory as atlas.py
    zip_filepath = filepath + ".zip"

    if os.path.exists(filepath):
        return  # No need to download

    if os.path.exists(zip_filepath):
        print(f"Zip file '{filename}.zip' already exists. Skipping Download.")
    else:
        confirm = input(f"Atlas model (3GB) needs to be downloaded. Do you want to download it? (yes/no): ")
        if confirm.lower() != "yes" and confirm.lower() != "y":
            print("Download cancelled.")
            return

        try:
            print(f"Downloading '{filename}' from {file_url}...")
            response = requests.get(file_url, stream=True)
            response.raise_for_status()  # Raise an exception for bad status codes (4xx or 5xx)

            with open(zip_filepath, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)

            print(f"Download of '{filename}' complete.")

        except requests.exceptions.RequestException as e:
            print(f"Error downloading '{filename}': {e}")
            if os.path.exists(zip_filepath): #remove the partially downloaded zip file
                os.remove(zip_filepath)
            return

    try:
        print(f"Unzipping '{filename}'...")
        with zipfile.ZipFile(zip_filepath, 'r') as zip_ref:
            zip_ref.extractall(os.path.dirname(filepath))  # Extract to the same directory
        print(f"Unzipping '{filename}' complete.")

        # Optionally, remove the zip file after extraction
        os.remove(zip_filepath)

    except zipfile.BadZipFile as e:
        print(f"Error unzipping '{filename}': {e}")
        if os.path.exists(filepath): #remove the potentially corrupted file
            os.remove(filepath)
    except Exception as e:
        print(f"An unexpected error occurred during unzip: {e}")
        if os.path.exists(filepath): #remove the potentially corrupted file
            os.remove(filepath)


def get_atlas_2d(embedding):
    """Generates 2D Atlas coordinates for traces."""
    download_and_unzip_atlas()
    reducer = joblib.load(FILEPATH)
    return reducer.transform(embedding)


def get_atlas_density_model():
    """Gets the density model for drawing Atlas boundaries."""
    return joblib.load(DENSITY_MODEL_FILEPATH)




"""Library for running model evaluation."""
import numpy as np
import scipy


def get_dwell_time(states, target_state, target_destination_state=None):
    """Calculates the dwell times for a trace."""
    
    current_dwell_time = 0
    last_state = None
    fret_dwell_time = []
    is_first_dwell = True
    
    for state in states:
        if last_state is None:
            last_state = state
            current_dwell_time = 0
            continue

        if state == last_state:
            current_dwell_time += 1
        else:
            if is_first_dwell:
                is_first_dwell = False
            else:
                if last_state == target_state:
                    if target_destination_state is None:
                        fret_dwell_time.append(current_dwell_time)
                    elif state == target_destination_state:
                        fret_dwell_time.append(current_dwell_time)
            current_dwell_time = 0
            last_state = state
    return fret_dwell_time


def get_cdf(data, bins):
    """Calculates CDF of a data."""
    cdf = []
    data = np.array(data)
    n = len(data)
    for b in bins:
        cdf.append(np.sum(data <= b))
    cdf = np.array(cdf) / n
    return cdf


def dwell_time_cdf(x, k):
    """Returns ideal CDF funciton for exponential distribution."""
    return 1 - np.exp(-k * x)


def estimate_k(tau, bins, time_resolution, tau_min=0, return_fit_err=False):
    """Estimates the mean and std of kinetic constant k given dwell times."""
    effective_tau = [s - tau_min for s in tau if s >= tau_min]
    tau_cdf = get_cdf(effective_tau, bins)
    popt, pcov = scipy.optimize.curve_fit(dwell_time_cdf, bins * time_resolution, tau_cdf)
    perr = np.sqrt(np.diag(pcov))
    k = popt[0]
    if return_fit_err:
        return (k, perr[0])
    else:
        return k


"""Library for the embedding projector."""
import os
import json
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from sciplotlib import style as spstyle
from smfret.dataset import FRETTraceSet

def trim(x, precision=4):
    """Trims a float to a given precision."""
    if int(x) == x:
        return x
    else:
        return float(np.format_float_positional(x, precision=4, unique=False, fractional=False, trim='k'))


def generate_sprite_plot(log_dir, trace_set):
    """Generates the .png file for sprite image of traces."""
    image_files = []
    with plt.style.context(spstyle.get_style('nature-reviews')):
        for idx in tqdm(range(trace_set.size)):
            fig = plt.figure(figsize=(2, 0.7))
            ax = fig.add_axes([0.2, 0.17, 0.68, 0.7])
            ax.plot(trace_set.traces[idx].acceptor, linewidth=0.1)
            ax.plot(trace_set.traces[idx].donor, linewidth=0.1)
            ax.set_xticks([])
            ax.set_yticks([])
            image_file = f'/tmp/trace_{idx}.png'
            image_files.append(image_file)
            fig.savefig(image_file, dpi=200, pad_inches=0, bbox_inches='tight', transparent=True)
            plt.close()

    
    images = []
    for f in image_files:
        e = Image.open(f)
        m = e.copy()
        e.close()
        images.append(m)
    width, height = images[0].size
    
    cols = int(np.sqrt(len(images))) + 1
    rows = cols
    
    total_width = cols * width
    total_height = rows * height
    
    new_im = Image.new('RGBA', (total_width, total_height))
    
    x_offset = 0
    for idx, im in enumerate(images):
      col = idx % cols
      row = idx // cols
      new_im.paste(im, (col * width, row * height))
    
    new_im.save(os.path.join(log_dir, 'trace_sprite.png'))


def generate_projector_data(log_dir, embedding, trace_set, plot_label, sprite=False):
    """Generates tensor and metadata for the embedding projector."""
    Path(log_dir).mkdir(parents=True, exist_ok=True)
    if sprite:
        generate_sprite_plot(log_dir, trace_set);
    with open(os.path.join(log_dir, 'metadata.tsv'), "w") as f:
      f.write('name\tground truth\tdonor\tacceptor\n')
      for index in range(embedding.shape[0]):
        donor = np.array([trim(x) for x in trace_set.traces[index].donor])
        acceptor = np.array([trim(x) for x in trace_set.traces[index].acceptor])
        f.write(f"trace {index}\t{plot_label[index]}\t{json.dumps(donor.tolist())}\t{json.dumps(acceptor.tolist())}\n")
    with open(log_dir + 'tensor.bytes', 'wb') as f:
        f.write(np.array(tf.cast(embedding, tf.float32)).tobytes())
    np.savetxt(log_dir + 'tensor.tsv', embedding, delimiter='\t')


def generate_tensor_and_metadata(trace_sets, start_frame, output_dir, encoder, max_frame=2000, max_traces=4000):
    """Generates the files for embedding projector."""

    embeddings = []
    labels = []
    colors = []
    color_count = 1
    image_files = []
    traces = []
    
    count = 0
    for trace_set in trace_sets:
        trim_size = len(trace_set.time) // 1000 * 1000
        trace_set.trim(trim_size)
        trace_set.trim(max_frame, start_frame=start_frame)
        trace_set.broadcast_data_to_traces()
        print(trace_set.size)
        if count + trace_set.size > max_traces:
            break
        else:
            count += trace_set.size
                
        with tf.device('/CPU:0'):
            embedding = encoder.predict(trace_set.to_tensor())
            embeddings.append(embedding)
            labels.append(np.max(trace_set.label, axis=-1))
            traces.extend(trace_set.traces)
    
    trace_set = FRETTraceSet()
    trace_set.traces = traces
    trace_set.size = len(traces)
    embedding = np.concatenate(embeddings, axis=0)
    labels = np.concatenate(labels, axis=0)
    generate_projector_data(output_dir, embedding, trace_set, labels, sprite=False)


from metasim.fret import core, tools
from metasim.fret.core import data
from metasim.fret.core.model import Model
from metasim.fret.core.score import get_entropy
from metasim.fret.tools import tuning
__all__ = ['core', 'tools']



"""Library for defining the learning tasks used in multi-task learning."""

import numpy as np
import tensorflow as tf
from typing import Callable

import smfret.dataset


class Task:
    """Base class for defining learning tasks."""
    feature_tensor: tf.Tensor = None
    labels:tf.Tensor = None
    def get_mini_batch(self, size):
        """Gets a mini batch of training data and labels."""
        ...
    def loss(self, y_pred, y_true) -> tf.Tensor:
        """Calculates the loss given input data and labels."""
        ...
    def predict(self, input_tensor) -> tf.Tensor:
        """Runs the inference given the input data."""
        ...
        
class SimpleTask(Task):
    """A simple task class that only envolve one feature tensor and one label tensor."""
    def __init__(self, feature_tensor, labels, model=None, name=None, shuffle=True):
        self.name = name  # A human readable name for readability
        self.num_traces = feature_tensor.shape[0]
        self.feature_tensor = feature_tensor
        self.labels = labels
        self.model = model
        self.shuffle = shuffle
        self.batch_counter = 0
        self.epochs = 0
        self.run_shuffle()

    def run_shuffle(self):
        """Runs the shuffle for the order of how a mini batch of data is read."""
        if self.shuffle:
            self.shuffle_order = np.random.permutation(list(range(self.num_traces)))
        else:
            self.shuffle_order = list(range(self.num_traces))

    def get_mini_batch(self, size):
        """Gets a mini batch of data."""
        n = self.shuffle_order[self.batch_counter]
        self.batch_counter += 1
        if self.batch_counter >= self.num_traces:
            self.batch_counter = 0
            self.epochs += 1
            self.run_shuffle()
            
        if n + size > self.num_traces:  # Exceeds the upper bound of the feature tensor
            return self.get_mini_batch(size)  # Finds the next valid return

        batch_tensor = self.feature_tensor[n : n + size, ...]
        batch_labels = self.labels[n : n + size, ...]
        
        return (batch_tensor, batch_labels)
    
    def train(self, epochs=1, **kwargs):
        """Trains the model."""
        return self.model.fit(x=self.feature_tensor, y=self.labels, shuffle=self.shuffle, epochs=epochs, **kwargs)
        
        


"""Library for model-based scores defined in the META-SiM paper."""
import numpy as np
import tensorflow as tf


NN_K = 50  # how many nearest neighbors are considered for the score


def calculate_distance(embedding):
    """Calculates the distance matrix from the embeddings."""
    with tf.device('/CPU:0'):
        distance = (
            tf.expand_dims(tf.einsum('ik,ik->i', embedding, embedding), axis=-1)
            + tf.expand_dims(tf.einsum('ik,ik->i', embedding, embedding), axis=0)
            - 2.0 * tf.einsum('ik,jk->ij', embedding, embedding))
    return distance


def get_entropy(embedding, label, branching_factor=NN_K):
    """Calculates the local Shannon entropy for traces.

    This function calculates the local Shannon entropy (LSE) defined in:
    https://doi.org/10.1101/2024.08.26.609721

    Args:
        embedding: The tensor of all embeddings.
        label: The list of all labels.
        branching_factor: The number of nearest neighbors used for calculation.

    Returns:
        A numpy array of entropy.
    """
    distance = calculate_distance(embedding)
    top_k_results = tf.math.top_k(
        -distance, k=branching_factor + 1, sorted=True, name=None
    )
    top_k_idx = top_k_results.indices.numpy()
    n = embedding.shape[0]
    entropy = []
    for i in range(n):
        this_label = label[i]
        neighbour_label = label[top_k_idx[i, :].flatten()]
        unique_labels = set(neighbour_label)
        shannon_entropy = 0
        for l in unique_labels:
            p = np.mean(neighbour_label == l)
            shannon_entropy -= p * np.log(p)
        entropy.append(shannon_entropy)
    return np.array(entropy)


def get_self_consistency_score(embedding, label, target_label, branching_factor=NN_K):
    """Calculates the self-consistency score (LCS) defined in the META-SiM paper.

    This function calculated the self-consistency score (LCS) for labels, defined in:
    https://doi.org/10.1101/2024.08.26.609721

    LCS is a score for each unique value of label, and can be further averaged across
    all unique values of the labels.

    Args:
        embedding: The full embedding tensor for all traces.
        label: The full list of labels for all traces.
        target_label: The label for which the score is calculated.
        branching_factor: The number of nearest neighbors used for calculation.

    Returns:
        A numpy array of LCS values.
    """

    distance = calculate_distance(embedding)
    top_k_results = tf.math.top_k(
        -distance, k=branching_factor+1, sorted=True, name=None
    )
    top_k_idx = top_k_results.indices.numpy()
    selected_labels = (label == target_label)
    top_k_labels = label[top_k_idx[selected_labels, 1:].flatten()]
    p_e = np.mean(label == target_label)
    return np.mean(top_k_labels == target_label)



"""Simulates smFRET traces."""

import numpy as np
import dataclasses
from typing import Callable
import logging

import smfret.dataset


DEFAULT_RNG = np.random.default_rng()

@dataclasses.dataclass
class SimulatorParameter:
    """Data class for parameters used in the simulator."""
    trace_length: int  # The total length of the generated trace
    num_states: int  # Number of FRET states
    fret_states: np.ndarray  # FRET states
    transition_prob_matrix: np.ndarray  # The transition probability matrix between FRET states
    initial_state: int  # The index of the initial state
    donor_lifetime: float  # Average donor lifetime before photobleaching
    acceptor_lifetime: float  # Average acceptor lifetime before photobleaching
    non_cy3_blink_lifetime: float  # Average contiguous trace length before a cy3 blinking
    cy3_blink_lifetime: float  # Average lengthof a cy3 blinking event
    total_intensity: float  # True total intensity of the smFRET trace
    background: float  # Average intensity of background.
    snr_background: float  # (total intensity + background) / std of background noise
    snr_signal: float  # (total intensity + background) / std of signal noise
    quantum_yield: float  # observed acceptor intensity / true acceptor intensity
    
    @property
    def donor_photobleach_prob(self):
        """Converts donor lifetime into donor photobleaching probability."""
        return 1.0 / self.donor_lifetime
    
    @property
    def acceptor_photobleach_prob(self):
        """Converts acceptor lifetime into acceptor photobleaching probability."""
        return 1.0 / self.acceptor_lifetime
    
    @property
    def cy3_blink_probability(self):
        """Converts Cy3 blink lifetime into Cy3 blink probability."""
        return 1.0 / self.non_cy3_blink_lifetime
    
    @property
    def donor_states(self):
        """Converts FRET states into donor intensities."""
        return [self.total_intensity * (1.0 - fret) for fret in self.fret_states] 
    
    @property
    def acceptor_states(self):
        """Converts FRET states into acceptor intensities."""
        return [self.total_intensity * (fret) for fret in self.fret_states] 
    
    @property
    def background_std(self):
        """Converts signal-to-noise ratio to std of background noise."""
        return (self.total_intensity + self.background) / self.snr_background
    
    @property
    def signal_std(self):
        """Converts SNR of signal to the std of signal."""
        return (self.total_intensity + self.background) / self.snr_signal
    

@dataclasses.dataclass
class ParameterGenerator:
    """Class for generating SimulatorParameter with given distributions."""
    trace_length_fn: Callable = lambda: 2000
    num_states_fn: Callable = lambda: DEFAULT_RNG.integers(low=1, high=5)
    fret_states_fn: Callable = lambda: DEFAULT_RNG.uniform(low=0.1, high=0.9)
    transition_prob_fn: Callable = lambda: DEFAULT_RNG.gamma(shape=5, scale=0.1)
    initial_state_fn: Callable = lambda fret_states: np.argmin(fret_states)
    donor_lifetime_fn: Callable = lambda: DEFAULT_RNG.uniform(low=800, high=1600)
    acceptor_lifetime_fn: Callable = lambda: DEFAULT_RNG.uniform(low=800, high=1600)
    non_cy3_blink_lifetime_fn: Callable = lambda: DEFAULT_RNG.uniform(low=200, high=1000)
    cy3_blink_lifetime_fn: Callable = lambda: DEFAULT_RNG.uniform(low=5, high=20)
    total_intensity_fn: Callable = lambda: DEFAULT_RNG.uniform(low=500, high=1500)
    background_fn: Callable = lambda: DEFAULT_RNG.uniform(low=10, high=100)
    snr_background_fn: Callable = lambda: DEFAULT_RNG.uniform(low=8, high=16)
    snr_signal_fn: Callable = lambda: DEFAULT_RNG.uniform(low=4, high=8)
    quantum_yield_fn: Callable = lambda: DEFAULT_RNG.uniform(low=0.95, high=1.0)
    
    def generate(self) -> SimulatorParameter:
        """Generates a SimulatorParameter object."""
        num_states = self.num_states_fn()
        fret_states = [self.fret_states_fn() for _ in range(num_states)]
        transition_prob_matrix = np.array(
            [[self.transition_prob_fn() for _ in range(num_states)] for _ in range(num_states)]
        )
        for i in range(num_states):
            transition_prob_matrix[i, i] += 1.0 - transition_prob_matrix[i, :].sum()
                
        return SimulatorParameter(
            trace_length=self.trace_length_fn(),
            num_states=num_states,
            fret_states=fret_states,
            transition_prob_matrix=transition_prob_matrix,
            initial_state=self.initial_state_fn(fret_states),
            donor_lifetime=self.donor_lifetime_fn(),
            acceptor_lifetime=self.acceptor_lifetime_fn(),
            non_cy3_blink_lifetime=self.non_cy3_blink_lifetime_fn(),
            cy3_blink_lifetime=self.cy3_blink_lifetime_fn(),
            total_intensity=self.total_intensity_fn(),
            background=self.background_fn(),
            snr_background=self.snr_background_fn(),
            snr_signal=self.snr_signal_fn(),
            quantum_yield=self.quantum_yield_fn(),
        )

    def copy(self):
        """Returns a copy of the same object."""
        return ParameterGenerator(dataclasses.asdict(self))
    
class SimulatedFRETTrace(smfret.dataset.FRETTrace):
    """Class for a simulated smFRET trace."""
    EPS = 1e-6
    def __init__(self, donor, acceptor, time, label, donor_ideal=None, acceptor_ideal=None, donor_lifetime=None, acceptor_lifetime=None):
        super().__init__(
            donor=donor,
            acceptor=acceptor,
            time=time,
            label=label,
        )
        self.donor_ideal = donor_ideal
        self.acceptor_ideal = acceptor_ideal
        self.donor_lifetime = donor_lifetime
        self.acceptor_lifetime = acceptor_lifetime
    
    @property
    def fret_ideal(self):
        """Ideal FRET values without noise."""
        return (
            self.acceptor_ideal / 
            (self.EPS + self.donor_ideal + self.acceptor_ideal)
        )
    
    @property 
    def total_ideal(self):
        """Ideal total intensity."""
        return self.donor_ideal + self.acceptor_ideal
        
    
class Simulator:
    """Base class for running smFRET simulations."""
    def __init__(self, parameter: SimulatorParameter):
        self.parameter = parameter
        self.rng = np.random.default_rng()
    
    def generate(self):
        """Generates a smFRET trace."""
        # Generic HMM simulation excluding Cy3 blinking and photobleaching
        size = self.parameter.trace_length
        n_states = self.parameter.num_states
        parameter = self.parameter
        state = np.zeros(size, dtype=np.int64)
        label = np.ones(size, dtype=np.int64)
        for i in range(size):
            if i == 0:
                state[i] = parameter.initial_state
            else:
                state[i] = self.rng.choice(
                    range(n_states),
                    p=self.parameter.transition_prob_matrix[state[i - 1], :],
                )
        donor = np.take(np.array(self.parameter.donor_states), state)
        acceptor = np.take(np.array(self.parameter.acceptor_states), state)
        
        # Store ideal donor and acceptor intensities
        donor_ideal = np.array(donor)
        acceptor_ideal = np.array(acceptor)
        
        # Quantum yield effects
        acceptor *= parameter.quantum_yield
        
        # Noise effects for signals
        effective_std = np.sqrt(
            max(0, parameter.signal_std**2 - parameter.background_std**2)
        )
        donor_noise_signal =  self.rng.normal(
            loc=0, scale=(np.sqrt(0.5) * effective_std), size=size)
        donor += donor_noise_signal  # Equal std in two channels is a good simple assumption
        acceptor_noise_signal = self.rng.normal(
            loc=0, scale=(np.sqrt(0.5) * effective_std), size=size)
        acceptor += acceptor_noise_signal
        
        # Photobleach effects
        donor_bleach_time = np.int64(self.rng.exponential(scale=parameter.donor_lifetime))
        acceptor_bleach_time = np.int64(self.rng.exponential(scale=parameter.acceptor_lifetime))
        if acceptor_bleach_time < size:
            donor[acceptor_bleach_time:] += acceptor[acceptor_bleach_time:]
            acceptor[acceptor_bleach_time:] = 0
            label[acceptor_bleach_time:] = 0
            donor_ideal[acceptor_bleach_time:] += acceptor_ideal[acceptor_bleach_time:]
            acceptor_ideal[acceptor_bleach_time:] = 0
        if donor_bleach_time < size:
            donor[donor_bleach_time:] = 0
            acceptor[donor_bleach_time:] = 0
            label[donor_bleach_time:] = 0
            donor_ideal[donor_bleach_time:] = 0
            acceptor_ideal[donor_bleach_time:] = 0
        
        
        # Cy3 blinking effects
        effective_size = min(donor_bleach_time, size)
        blink_frame = np.int64(self.rng.exponential(scale=parameter.non_cy3_blink_lifetime))
        while blink_frame < effective_size:
            blink_lifetime = np.int64(self.rng.exponential(scale=parameter.cy3_blink_lifetime))
            if blink_lifetime > 0:
                blink_end_frame = min(effective_size, blink_frame + blink_lifetime)
                donor[blink_frame:blink_end_frame] = 0
                acceptor[blink_frame:blink_end_frame] = 0
                label[blink_frame:blink_end_frame] = 0
                donor_ideal[blink_frame:blink_end_frame] = 0
                acceptor_ideal[blink_frame:blink_end_frame] = 0
            blink_frame += blink_lifetime
            blink_frame += np.int64(self.rng.exponential(scale=parameter.non_cy3_blink_lifetime))
        
        # Noise effects for background
        donor_noise_background = self.rng.normal(
            loc=0.5 * parameter.background,
            scale=(np.sqrt(0.5) * parameter.background_std),
            size=size,
        )   # Equal std in two channels is a good simple assumption
        donor += donor_noise_background 
        acceptor_noise_background = self.rng.normal(
            loc=0.5 * parameter.background,
            scale=(np.sqrt(0.5) * parameter.background_std),
            size=size,
        )
        acceptor += acceptor_noise_background
        
        donor_bleach_time = min(donor_bleach_time, size)
        acceptor_bleach_time = min(acceptor_bleach_time, donor_bleach_time)
        
        time = np.array(list(range(1, size + 1)))
        trace = SimulatedFRETTrace(
            donor=donor,
            acceptor=acceptor,
            time=time,
            label=label,
            donor_ideal=donor_ideal,
            acceptor_ideal=acceptor_ideal,
            donor_lifetime=donor_bleach_time,
            acceptor_lifetime=acceptor_bleach_time,
        )
        return trace


class SingleChannelSimulator(Simulator):
    """Class for a simulator of single channel data."""
    EPS = 1e-7
    def generate(self):
        """overrides the two-channel generating logic."""
        trace = super().generate()
        trace.donor *= self.EPS
        return trace


class SimulatedTraceSet(smfret.dataset.FRETTraceSet):
    """Class for storing simulated smFRET traces."""
    def __init__(self, simulator: Simulator):
        super().__init__()
        self.simulator = simulator
    
    def populate(self, n: int):
        """Populates the dataset using the simulator."""
        self.size = n
        self.traces = [self.simulator.generate() for _ in range(n)]
        self.time = self.traces[0].time
        self.donor = np.stack([trace.donor for trace in self.traces], axis=0)
        self.acceptor = np.stack([trace.acceptor for trace in self.traces], axis=0)
        self.label = np.stack([trace.label for trace in self.traces], axis=0)
        self.is_labeled = True


"""Library for FRET idealization using META-SiM."""

import os
import numpy as np
import metasim.fret.core.data as core_data
import metasim.fret.core.model as core_model


CHECKPOINT_PATH = 'checkpoints/model-20240511-131505/FRETSTATE.h5'
CHECKPOINT_PATH = os.path.join(os.path.dirname(__file__), CHECKPOINT_PATH)
MODEL_RESOLUTION = 0.1


def idealize(dataset):
    """Generates idealized FRET intensities for a dataset."""
    model = core_model.Model(checkpoint_path=CHECKPOINT_PATH)
    pred = model.encode_frames(dataset)
    idealized_states = [
        np.argmax(p, axis=-1) * MODEL_RESOLUTION for p in pred
    ]
    # Reduce half of resolution for non-zero FRET states to offset
    # quantization error
    idealized_states = [
        p - (p > 0) * 0.5 * MODEL_RESOLUTION for p in idealized_states
    ]
    return idealized_states


def get_fret_efficiency(dataset, idealized_states):
    """Gets the list of FRET efficiency values.

    In this function, we use state idealization to identify active FRET regions.
    Activate region is defined as state > 0, per training definition.

    Args:
        dataset: A dataset object.
        idealized_states: A list of numpy array of FRET state.

    Returns:
        A list of numpy array of FRET efficiencies. This list can be used to construct FRET histogram.
    """
    tensors = dataset.to_tensors()
    efficiency = []
    for i, tensor in enumerate(tensors):
        states = idealized_states[i]
        fret = tensor[..., 1] / (tensor[..., 0] + tensor[..., 1])
        values = fret[states > 0]
        efficiency.append(np.array(values))
    return efficiency

